// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus
#include <boost/program_options.hpp>
#include <cstdint>
#include <iostream>	 // for basic_ostream::operator<<
#include <thread>

#include "BinaryPrivateSemaphore.hpp"		 // for BPS
#include "CountingPrivateSemaphore.hpp"	 // for CPS
#include "Debug.hpp"										 // for DBG
#include "Fiber.hpp"										 // for Fiber
#include "PrivateSemaphore.hpp"					 // for PS
#include "Runtime.hpp"									 // for Runtime
#include "lib/sync/Semaphore.hpp"

namespace po = boost::program_options;

using fibParams = struct {
	uint64_t n;
	uint64_t* result;
	PS* sem;
};

static void fib(void* voidParams) {
	auto* params = static_cast<fibParams*>(voidParams);
	uint64_t n = params->n;
	auto* result = params->result;
	PS* sem = params->sem;

	if (n < 2) {
		*result = n;
	} else {
		CPS newSem(2);

		uint64_t a, b;

		fibParams newParams1;
		newParams1.n = n - 1;
		newParams1.result = &a;
		newParams1.sem = &newSem;

		// Note that this is the inefficient spawn/sync variant, we
		// usually would compute one previous fib number without spawning.
		fibParams newParams2;
		newParams2.n = n - 2;
		newParams2.result = &b;
		newParams2.sem = &newSem;

		Fiber* f1 = Fiber::from(&fib, &newParams1);
		Fiber* f2 = Fiber::from(&fib, &newParams2);

		Runtime* runtime = Runtime::getRuntime();
		runtime->schedule(*f1);
		runtime->schedule(*f2);

		DBG("fib: Calling wait for n=" << n);
		newSem.wait();

		*result = a + b;
	}

	DBG("fib: Calling signalAndExit for n=" << n);
	sem->signalAndExit();
}

// NOLINTNEXTLINE(bugprone-exception-escape)
auto main(int argc, char* argv[]) -> int {
	uint64_t fibNum = 12;
	po::options_description desc("Allowed options");
	// clang-format off
	desc.add_options()
		("help", "Show help")
		("nthreads", po::value<unsigned int>()->default_value(std::thread::hardware_concurrency()), "Number of worker threads used by EMPER's runtime system")
		("fibnum", po::value<uint64_t>(&fibNum)->default_value(fibNum), "The Fibonacci number to compute")
		;
	// clang-format on

	// Make 'fibnum' a positional option.
	po::positional_options_description pos_desc;
	pos_desc.add("fibnum", -1);

	// clang-format off
	auto parse_result = po::command_line_parser(argc, argv)
		.options(desc)
		.positional(pos_desc)
		.run()
		;
	// clang-format on

	po::variables_map vm;
	po::store(parse_result, vm);
	po::notify(vm);

	const unsigned nthreads = vm["nthreads"].as<unsigned int>();

	std::cout << "Number of threads: " << nthreads << std::endl;

	Runtime runtime(nthreads);

	emper::lib::sync::Semaphore semaphore;

	Fiber* fibFiber = Fiber::from([&] {
		uint64_t result;
		BPS sem;
		fibParams params = {fibNum, &result, &sem};

		fib(&params);

		sem.wait();

		std::cout << "fib(" << fibNum << ") = " << result << std::endl;

		semaphore.notify();
	});

	runtime.scheduleFromAnywhere(*fibFiber);

	semaphore.wait();

	return 0;
}
