// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus
#include "Context.hpp"

#include <sys/mman.h>

#include <cerrno>
#include <cstdint>
#include <ostream>

#include "stats/Worker.hpp"

thread_local Context* Context::currentContext;

auto Context::calcTos(char* context) -> void* {
	auto contextBase = (uintptr_t)context;
	uintptr_t contextTop = contextBase + CONTEXT_SIZE;
	// Align the Top-of-Stack (tos) to 16 byte.
	uintptr_t tos = contextTop & ~0xf;
	uintptr_t tosOffset = tos - contextBase;
	return context + tosOffset;
}
auto Context::calcTos() -> void* { return calcTos(context); }

auto Context::calcGuardPage(char* context) -> void* {
	if constexpr (!emper::STACK_GUARD_PAGE) return nullptr;

	auto contextBase = (uintptr_t)context;
	uintptr_t guardPageBase = em ::roundUp(contextBase, PAGE_SIZE);
	uintptr_t guardPageOffset = guardPageBase - contextBase;
	return context + guardPageOffset;
}
auto Context::calcGuardPage() -> void* { return calcGuardPage(context); }

auto Context::inGuardPage(void* mem) -> bool {
	if constexpr (!emper::STACK_GUARD_PAGE) return false;

	const auto _guardPage = (uintptr_t)this->guardPage;
	const auto _mem = (uintptr_t)mem;
	return _mem >= _guardPage && _mem < _guardPage + PAGE_SIZE;
}
auto Context::inCurrentGuardPage(void* mem) -> bool {
	if (!currentContext) return false;
	return currentContext->inGuardPage(mem);
}

auto Context::calcBos(char* context) -> void* {
	auto contextBase = (uintptr_t)context;
	uintptr_t bosOffset = 0;
	if constexpr (emper::STACK_GUARD_PAGE) {
		uintptr_t guardPageBase = em ::roundUp(contextBase, PAGE_SIZE);
		uintptr_t guardPageTop = guardPageBase + PAGE_SIZE;
		bosOffset = guardPageTop - contextBase;
	}
	return context + bosOffset;
}
auto Context::calcBos() -> void* { return calcBos(context); }

// cppcheck-suppress noExplicitConstructor selfInitialization
Context::Context(func_t mainFunction)
		: tos(calcTos()),
			guardPage(calcGuardPage()),
			bos(calcBos()),
			usableStackSize((char*)tos - (char*)bos),
			mainFunction(std::move(std::move(mainFunction))) {
	//		valgrindStackId = VALGRIND_STACK_REGISTER(context, context + CONTEXT_SI);

	if constexpr (emper::STACK_GUARD_PAGE) {
		int res = mprotect(guardPage, PAGE_SIZE, PROT_NONE);
		if (res) DIE_MSG_ERRNO("Installaing guard page at " << guardPage << " failed");
	}

	maybeMarkStack();

	setEmptyHook();

	// TOS is 16-byte aligned, however we need the initial
	// savedStackpointer to be 8-byte aligned because will later jmp to
	// it. We also save a pointer to very first function this context is
	// going to execute a this initial stackpointer value.
	alphaFunctionIpLocation = savedStackpointer = (uintptr_t*)tos - 1;

	void** alphaSavedIp = reinterpret_cast<void**>(savedStackpointer);
	void (*f)() = &kickoff;
	assert(f != nullptr);
	*(alphaSavedIp) = reinterpret_cast<void*>(f);
}

Context::~Context() {
	//		VALGRIND_STACK_DEREGISTER(valgrindStackId);
	if constexpr (emper::STACK_GUARD_PAGE) {
		int res = mprotect(guardPage, PAGE_SIZE, PROT_READ | PROT_WRITE | PROT_EXEC);
		if (res) DIE_MSG_ERRNO("Uninstallaing guard page at " << guardPage << " failed");
	}
}

void Context::printTo(std::ostream& strm, bool withPtr) const {
	// clang-format off
	strm << "Context ";
	if (withPtr) {
		strm << "(" << this << ") ";
	}

	strm << "[fiber=" << currentFiber
       << ", tos=" << tos;
	strm << ", bos=" << &context
			 << ", saved-sp=" << savedStackpointer
			 << ", *saved-sp=" << *static_cast<void**>(savedStackpointer);
	if (savedStackpointer != alphaFunctionIpLocation) {
	strm << ", alpha-fun-loc=" << alphaFunctionIpLocation
		   << ", *alpha-func-loc=" << *static_cast<const void* const*>(alphaFunctionIpLocation);
	}
	strm << "]";
	// clang-format off
}

auto operator<<(std::ostream& strm, const Context& context) -> std::ostream& {
	context.printTo(strm, false);
	return strm;
}

auto operator<<=(std::ostream& strm, const Context& context) -> std::ostream& {
	context.printTo(strm, true);
	return strm;
}

void Context::unmap(void* sp) const {
	constexpr size_t PAGE_SIZE_MASK = PAGE_SIZE - 1;
	const auto spBase = (uintptr_t)sp;
	const uintptr_t spBasePage = spBase & ~PAGE_SIZE_MASK;
	const auto bosBase = (uintptr_t)bos;
	const uintptr_t bosBasePage = (bosBase + PAGE_SIZE_MASK) & ~PAGE_SIZE_MASK;
	const size_t length = spBasePage - bosBasePage;
	const uintptr_t bosBasePageOffset = bosBasePage - bosBase;
	auto* const addr = (char*)bos + bosBasePageOffset;
	const int advice = []{
		if constexpr (emper::CONTINUATION_STEALING_MADVISE_STACK == emper::ContinuationStealingMadviseStack::dontneed) {
			return MADV_DONTNEED;
		} else {
			return MADV_FREE;
		}
	}();

	LOGD("madvise() addr=" << addr << ", length=" << length << ", advice=" << advice);

	errno = madvise(addr, length, advice);
	if (errno) {
		DIE_MSG_ERRNO("Unmapping unused stack space failed");
	}
}

void Context::maybeMarkStack() {
	if constexpr (!emper::DEBUG && !emper::STATS_STACK_USAGE_ENABLED) return;

	// Write the stack full of 0xcc bytes, which just happen to be
	// the 'int3' instruction, which will trigger a breakpoint
	// when executed.
	void* res = memset(bos, 0xcc, usableStackSize);
	if (!res) DIE;

	// Mark the last valid 16 bytes just below the top of stack.
	res = memset(((uintptr_t*)tos) - 1, 0xab, sizeof(uintptr_t));
	if (!res) DIE;

	// Mark the eventually existing unused top stack bytes
	res = memset(tos, 0xee, &context[CONTEXT_SIZE] - (char*)tos);
	if (!res) DIE;
}

void Context::recordStackUsage() {
	constexpr uintptr_t stackMarker = 0xcccccccccccccccc;

	size_t bosOffset;
	for (bosOffset = 0; bosOffset < usableStackSize; bosOffset += sizeof(stackMarker)) {
		auto* stackWord = (uintptr_t*)((char*)bos + bosOffset);
		if (*stackWord != stackMarker) break;
	}

	size_t stackUsage = usableStackSize - bosOffset;

	emper::stats::worker->recordStackUsage(stackUsage);
}
