// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus
#include "ContextManager.hpp"

#include <cassert>	// for assert
#include <new>
#include <ostream>
#include <utility>

#include "Context.hpp"
#include "Continuation.hpp"
#include "Debug.hpp"
#include "Dispatcher.hpp"
#include "Emper.hpp"
#include "Fibril.hpp"
#include "Runtime.hpp"
#include "emper-config.h"	 // // IWYU pragma: keep

class AbstractFiber;

ContextManager::ContextManager(Runtime& runtime) : MemoryManager(runtime), runtime(runtime) {}

auto ContextManager::getFreeContext() -> Context* {
	if constexpr (!emper::CONTEXT_MANAGER_WITH_MEMORY_MANAGER) {
		return new Context(this->runtime.dispatcher.getDispatchLoop());
	}

	bool malloced;
	void* memory = getMemory(&malloced);
	if (malloced) {
		return new (memory) Context(this->runtime.dispatcher.getDispatchLoop());
	}

	auto* context = static_cast<Context*>(memory);
	context->reinit();
	if constexpr (emper::STATS_STACK_USAGE == emper::StatsStackUsage::accurate) {
		// If we want accurate stack-usage stats, then we need to re-mark
		// the stack if we got the memory from the memory manager but if
		// it wasn't malloced.
		context->maybeMarkStack();
	}
	return context;
}

void ContextManager::putFreeContext(Context* context) {
	if constexpr (emper::STATS_STACK_USAGE_ENABLED) {
		context->recordStackUsage();
	}

	if constexpr (emper::CONTEXT_MANAGER_WITH_MEMORY_MANAGER) {
		putMemory(context);
	} else {
		delete context;
	}
}

thread_local static Continuation* continuation;

void ContextManager::start() {
	Continuation originalStackContinuation;

	continuation = &originalStackContinuation;

	uintptr_t fibrilResumeValue = originalStackContinuation.setJmp();

	// We perform a longjmp with fibrilResumeValue == 0 to switch back
	// to the original stack and terminate the runtime.
	if (unlikely(!fibrilResumeValue && runtime.isTerminating())) {
		Context* currentContext = Context::currentContext;
		if (currentContext) [[likely]] {
			Context::currentContext = nullptr;
			putFreeContext(currentContext);
		}
		return;
	}

	Context* freeContext = nullptr;
	if constexpr (emper::CONTINUATION_STEALING) {
		LOGD("Fibril::tryResumeFibril(" << fibrilResumeValue << ")");

		Fibril::tryResumeFibril(fibrilResumeValue);

		// Resuming the fibril was not (yet) possible. There may be a
		// current context, even though we are currently here on the
		// initial (pthread) context. Lookup the current context and
		// reinitilize it so it can be re-used to start() it and continue
		// in the dispatch loop.
		freeContext = Context::getCurrentContext();
	}

	if (freeContext) {
		freeContext->reinit();
	} else {
		freeContext = getFreeContext();
	}

	freeContext->start();
}

[[noreturn]] void ContextManager::switchToOriginalStack() {
	DBG("Switching to original stack from " << Context::currentContext);
	continuation->longJmp(0);
}

void ContextManager::tryResumeFibril(uintptr_t fibrilResumeValue) {
	continuation->longJmp(fibrilResumeValue);
}

/**
 * Save the current context and start a new one.
 */
void ContextManager::saveAndStartNew(func_t freshContextHook) {
	Context* freeContext = getFreeContext();
	freeContext->setHook(std::move(freshContextHook));
	Context::getCurrentContext()->saveAndStart(freeContext);
}

#if Debug > EMPER_LOG_LEVEL
#define HOOK_CAPTURE_LIST contextToFree
#else
// We only need to caputure 'this' for the LOGD() debug macro.
#define HOOK_CAPTURE_LIST this, contextToFree
#endif

/**
 * Discard the current context and resume a saved one.
 */
// The discardAndResume() method could theoretically be made static in
// non-debug builds.
// NOLINTNEXTLINE(readability-convert-member-functions-to-static)
void ContextManager::discardAndResume(Context* context) {
	assert(context != nullptr);

	Context* contextToFree = Context::getCurrentContext();
	context->setHook([HOOK_CAPTURE_LIST] {
		LOGD("Freeing context " << contextToFree);
		putFreeContext(contextToFree);
	});

	// Since we are going to discard this context, it will never reach
	// the end of its dispatch loop, and hence we need to ensure that
	// its fiber is recycled.
	AbstractFiber* currentFiber = contextToFree->getFiber();
	assert(currentFiber);
	LOGD("Recycling fiber " << currentFiber << " of discard context " << contextToFree);
	runtime.dispatcher.recycle(currentFiber);

	contextToFree->discardAndResume(context);
}
