// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus
#pragma once

#include <ostream>

#include "AbstractFiber.hpp"
#include "Common.hpp"
#include "Context.hpp"
#include "Debug.hpp"
#include "Emper.hpp"
#include "Fiber.hpp"
#include "emper-common.h"

class Runtime;
class ContextManager;

class Dispatcher : public Logger<LogSubsystem::DISP> {
 private:
	void dispatchLoopSleep();

	/**
	 * Marker bool if the worker dispatched work, used for "worker sleep" stats.
	 */
	static thread_local bool dispatchedWork;

 protected:
	Runtime& runtime;

	void dispatchLoopDoSleep();

	virtual void dispatchLoop() = 0;

	auto getDispatchLoop() -> func_t;

	// The dispatch() method could theoretically be made static in
	// non-debug builds.
	// NOLINTNEXTLINE(readability-convert-member-functions-to-static)
	inline void dispatch(AbstractFiber* fiber) {
		LOGD("executing " <<= *fiber);

		if constexpr (emper::STATS_WORKER_SLEEP) {
			dispatchedWork = true;
		}

		Context::setCurrentFiber(fiber);
		fiber->run();
	}

	static inline auto isRunnable(AbstractFiber* abstractFiber) -> bool {
		auto* fiber = abstractFiber->asFiberIfPossible();
		if (!fiber) return true;

		if (fiber->isMultiFiber()) {
			return fiber->setRunnableFalse();
		}
		return true;
	}

	static inline auto getAffinityBuffer(AbstractFiber* fiber) -> workeraffinity_t* {
		return fiber->getAffinityBuffer();
	}

	static inline auto decreaseRefCount(Fiber* fiber) -> unsigned int {
		return fiber->doAtomicDecrRefCount();
	}

	void recycle(AbstractFiber* abstractFiber) {
		auto* fiber = abstractFiber->asFiberIfPossible();

		// We only recycle Fibers.
		if (!fiber) {
			LOGD("not recyling non-fiber " << fiber)
			return;
		}

		recycle(fiber);
	}

	virtual void recycle(Fiber* fiber) {
		LOGD("recyling fiber " << fiber)
		delete fiber;
	}

 public:
	Dispatcher(Runtime& runtime) : runtime(runtime) {}

	virtual ~Dispatcher() = default;

	static auto getCurrentFiber() -> AbstractFiber& {
		AbstractFiber* fiber = getCurrentFiberPtr();
		return *fiber;
	}

	static auto getCurrentFiberPtr() -> AbstractFiber* { return Context::getCurrentFiber(); }

	static auto isDispatchedControlFlow() -> bool { return getCurrentFiberPtr() != nullptr; }

	friend ContextManager;
	friend class Scheduler;
};
