// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus
#pragma once

#include "Common.hpp"
#include "Emper.hpp"
#include "Runtime.hpp"
#include "emper-common.h"
#include "lib/adt/BoundedBumpArray.hpp"
#include "lib/adt/WsQueue.hpp"

template <typename T, intptr_t WS_QUEUE_SIZE, size_t WORKER_EXCLUSIVE_QUEUE_SIZE>
class MemoryManager {
 private:
	const workerid_t workerCount;

	adt::BoundedBumpArray<void, WORKER_EXCLUSIVE_QUEUE_SIZE>** workerExclusiveQueues;

	using WsQueue = lib::adt::wsqueue::Default<void*, WS_QUEUE_SIZE>;
	WsQueue** queues;

	static thread_local adt::BoundedBumpArray<void, WORKER_EXCLUSIVE_QUEUE_SIZE> workerExclusiveQueue;

	static thread_local WsQueue queue;

	static auto mallocMemory() -> void* {
		void* memory;
		int res = posix_memalign(&memory, alignof(T), sizeof(T));
		if (res != 0) {
			abort();
		}
		return memory;
	}

 public:
	MemoryManager(Runtime& runtime);

	~MemoryManager();

	auto getMemory(bool* malloced) -> void* {
		*malloced = false;
		void* memory = workerExclusiveQueue.get();
		if (memory) return memory;

		constexpr int ONE_RETRY = 1;
		auto stealingResult = queue.template popTop<ONE_RETRY>(&memory);

		if (likely(stealingResult == emper::StealingResult::Stolen)) return memory;

		// Fallback to memory-stealing.
		const workerid_t myWorkerId = Runtime::getWorkerId();
		const workerid_t startWorkerId = Runtime::rand() % workerCount;
		const workerid_t abortStealingAt = workerCount * emper::MEMORY_MANAGER_VICTIM_PERCENTAGE;
		for (workerid_t i = 0; i < workerCount; ++i) {
			if (i == abortStealingAt) break;

			workerid_t victim = (startWorkerId + i) % workerCount;

			// Don't steal from ourselves.
			if (unlikely(victim == myWorkerId)) continue;

			constexpr int NO_RETRIES = 0;
			stealingResult = queues[victim]->template popTop<NO_RETRIES>(&memory);
			if (stealingResult == emper::StealingResult::Stolen) return memory;
		}

		*malloced = true;
		// If everything fails, allocate the memory.
		return mallocMemory();
	}

	void putMemory(T* memory) {
		assert(memory);
		bool pushed = workerExclusiveQueue.put(memory);
		if (pushed) return;

		pushed = queue.pushBottom(memory);
		if (pushed) return;

		free(memory);
	}
};

template <typename T, intptr_t WS_QUEUE_SIZE, size_t WORKER_EXCLUSIVE_QUEUE_SIZE>
thread_local adt::BoundedBumpArray<void, WORKER_EXCLUSIVE_QUEUE_SIZE>
		MemoryManager<T, WS_QUEUE_SIZE, WORKER_EXCLUSIVE_QUEUE_SIZE>::workerExclusiveQueue;

template <typename T, intptr_t WS_QUEUE_SIZE, size_t WORKER_EXCLUSIVE_QUEUE_SIZE>
thread_local lib::adt::wsqueue::Default<void*, WS_QUEUE_SIZE>
		MemoryManager<T, WS_QUEUE_SIZE, WORKER_EXCLUSIVE_QUEUE_SIZE>::queue;

template <typename T, intptr_t WS_QUEUE_SIZE, size_t WORKER_EXCLUSIVE_QUEUE_SIZE>
MemoryManager<T, WS_QUEUE_SIZE, WORKER_EXCLUSIVE_QUEUE_SIZE>::MemoryManager(Runtime& runtime)
		: workerCount(runtime.getWorkerCount()) {
	queues = new lib::adt::wsqueue::Default<void*, WS_QUEUE_SIZE>*[workerCount];
	workerExclusiveQueues =
			new adt::BoundedBumpArray<void, WORKER_EXCLUSIVE_QUEUE_SIZE>*[workerCount];

	auto newWorkerHook = [this](workerid_t workerId) {
		workerExclusiveQueues[workerId] = &workerExclusiveQueue;
		queues[workerId] = &queue;
	};
	runtime.addNewWorkerHook(newWorkerHook);
}

template <typename T, intptr_t WS_QUEUE_SIZE, size_t WORKER_EXCLUSIVE_QUEUE_SIZE>
MemoryManager<T, WS_QUEUE_SIZE, WORKER_EXCLUSIVE_QUEUE_SIZE>::~MemoryManager() {
	for (workerid_t i = 0; i < workerCount; ++i) {
		auto* queue = workerExclusiveQueues[i];
		while (auto* mem = queue->get()) free(mem);
	}
	delete workerExclusiveQueues;

	for (workerid_t i = 0; i < workerCount; ++i) {
		auto* queue = queues[i];
		void* mem;
		while (queue->popBottom(&mem)) free(mem);
	}
	delete queues;
}
