// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus
#pragma once

#include <cstddef>
#include <functional>	 // for function
#include <optional>
#include <ostream>

#include "CallerEnvironment.hpp"
#include "Debug.hpp"	// for LogSubsystem, LogSubsystem::SCHED, Logger
#include "Dispatcher.hpp"
#include "Emper.hpp"
#include "Fiber.hpp"
#include "FiberHint.hpp"
#include "emper-common.h"	 // for workeraffinity_t
#include "lib/adt/LockedUnboundedQueue.hpp"

class AbstractFiber;
class Runtime;
class RuntimeStrategy;
struct NextFiberResult;

class Scheduler : public Logger<LogSubsystem::SCHED> {
 private:
	lib::adt::LockedUnboundedQueue<Fiber> scheduleAnywhereQueue;

	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	void wakeupSleepingWorkers(emper::FiberHint hint);

	void notifyRuntimeAboutWorkStolen();

 protected:
	Runtime& runtime;
	Dispatcher& dispatcher;

	Scheduler(Runtime& runtime, RuntimeStrategy& strategy);

	virtual ~Scheduler() = default;

	void addNewWorkerHook(const std::function<void(workerid_t)>& hook);

	static inline auto getAffinityBuffer(Fiber& fiber) -> workeraffinity_t* {
		return fiber.getAffinityBuffer();
	}

	static inline void increaseRefCount(Fiber& fiber) { fiber.doAtomicIncrRefCount(); }

	template <CallerEnvironment callerEnvironment>
	inline void onNewWork(emper::FiberHint hint) {
		if constexpr (emper::WORKER_SLEEP) {
			wakeupSleepingWorkers<callerEnvironment>(hint);
		}
	}

	inline void onWorkStolen() {
		if constexpr (emper::WORKER_SLEEP) {
			notifyRuntimeAboutWorkStolen();
		}
	}

	void enqueueInAnywhereQueue(Fiber& fiber) { scheduleAnywhereQueue.enqueue(&fiber); }

	void insertInAnywhereQueue(Fiber** fibers, unsigned count) {
		scheduleAnywhereQueue.insert(fibers, count);
	}

	template <class OutputIt>
	auto dequeueFibersFromAnywhereQueue(OutputIt begin, OutputIt end) -> size_t {
		return scheduleAnywhereQueue.dequeue(begin, end);
	}

	virtual void scheduleInternal(Fiber& fiber) = 0;
	virtual void scheduleOnInternal(Fiber& fiber, workerid_t workerId) = 0;
	virtual void scheduleFromAnywhereInternal(Fiber& fiber) = 0;
	virtual void scheduleFromAnywhereInternal(Fiber** fibers, unsigned count) = 0;

	void recycle(AbstractFiber* fiber) { dispatcher.recycle(fiber); };

	virtual auto pushBottom(AbstractFiber& fiber) -> bool = 0;
	virtual auto popBottom() -> AbstractFiber* = 0;
	friend class Fibril;

	virtual auto nextFiber() -> std::optional<NextFiberResult> = 0;
	friend class Runtime;

 public:
	void schedule(Fiber& fiber) {
		// Calling schedule() only works from within the EMPER runtime.
		emper::assertInRuntime();

		LOGD("Scheduling fiber " << &fiber);
		scheduleInternal(fiber);
	}

	// TODO: maybe this should also be a specialized function
	void schedule(Fiber** fibers, unsigned count) {
		// Calling schedule() only works from within the EMPER runtime.
		emper::assertInRuntime();

		for (unsigned i = 0; i < count; ++i) {
			Fiber& fiber = *fibers[i];
			LOGD("Scheduling batched fiber " << &fiber);
			scheduleInternal(fiber);
		}
	}

	void scheduleOn(Fiber& fiber, workerid_t workerId);

	void scheduleFromAnywhere(Fiber& fiber) {
		LOGD("Scheduling fiber " << &fiber << " from anywhere");
		scheduleFromAnywhereInternal(fiber);
	}

	void scheduleFromAnywhere(Fiber** fibers, unsigned count) {
		scheduleFromAnywhereInternal(fibers, count);
	}
};
