// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021 Florian Fischer
#pragma once

#include <atomic>
#include <cassert>
#include <iostream>

#include "CallerEnvironment.hpp"
#include "Debug.hpp"
#include "Emper.hpp"
#include "emper-common.h"

namespace emper {

/**
 * @brief Class modeling the strategy used to wakeup suspended worker threads
 *
 * The possible behaviors are:
 * - 'one': Wakeup a SINGLE suspended worker for each onNewWork notification
 * - 'all': Wakeup ALL suspended worker for each onNewWork notification
 * - 'throttle': Only wakeup a suspended worker if there is currently no worker
 *               already waking up. The wakup strategy can be in a tri-state:
 *               (pending, notified, waking). 'Pending' means no unconsumed notification
 *               is available. 'Notified' means there is a consumable notification available.
 *               Waking means a worker has consumed the notification and entered
 *               its dispatch loop.
 *
 *               A waking worker will either pass its working status when it
 *               calls onNewWork itself, consumes a further notification from anywhere
 *               and skip sleeping to prevent lost wakups, or reset the state to pending when
 *               successfully stolen work.
 *               The 'throttle' strategy is heavily inspired by the zap thread-pool:
 *               https://zig.news/kprotty/resource-efficient-thread-pools-with-zig-3291
 */
class WakeupStrategy : public Logger<LogSubsystem::WAKE_S> {
	workerid_t workerCount;

	enum class ThrottleState {
		pending,
		notified,
		waking,
	};

	// TODO: figure out the weakest sound memory ordering of the atomic operations
	std::atomic<ThrottleState> throttleState = ThrottleState::pending;

	/**
	 * @brief Flag indicating that this worker is the waking thread
	 *
	 * When using the 'throttle' wakeup strategy the 'waking' thread is responsible
	 * to reset the notified flag of the wakeup strategy and thus allowing more
	 * threads to be awoken.
	 */
	static thread_local bool isWaking;

	auto canWakeupThrottle() -> bool {
		ThrottleState notified = ThrottleState::notified;
		bool wasNotified = throttleState.compare_exchange_strong(notified, ThrottleState::waking);

		// This is the waking thread now
		if (wasNotified) {
			isWaking = true;
		}

		LOGD("can " << (wasNotified ? "" : "not ") << "wakeup");

		return wasNotified;
	}

	template <CallerEnvironment callerEnvironment>
	auto getWakeupCountThrottle() -> workerid_t {
		bool wasPending;
		// If the waking worker calls onNewWork->getWakeupCount he is clearly done waking up
		if constexpr (callerEnvironment == CallerEnvironment::EMPER) {
			if (isWaking) {
				resetWaking();
				LOGD("waking worker wants to wakeup others -> 1");
				return 1;
			}

			ThrottleState state = ThrottleState::pending;
			wasPending = throttleState.compare_exchange_strong(state, ThrottleState::notified);

			LOGD("get wakeup count from " << callerEnvironment << " -> " << (wasPending ? 1 : 0));
			// If we notify from anywhere and there is currently a waking worker we have
			// to prevent it them from sleeping -> always set the state to notified;
		} else {
			wasPending = throttleState.exchange(ThrottleState::notified) == ThrottleState::pending;
		}

		return wasPending ? 1 : 0;
	}

	ThrottleState resetWaking() {
		assert(isWaking);
		isWaking = false;
		ThrottleState old = throttleState.exchange(ThrottleState::pending);
		return old;
	}

 public:
	WakeupStrategy(workerid_t workerCount) : workerCount(workerCount) {}

	void onWorkStolen() {
		if constexpr (emper::WORKER_WAKEUP_STRATEGY == emper::WorkerWakeupStrategy::throttle) {
			if (isWaking) {
				LOGD("waking worker stole work -> reset to pending");
				resetWaking();
			}
		}
	}

	bool canSleep() {
		if constexpr (emper::WORKER_WAKEUP_STRATEGY == emper::WorkerWakeupStrategy::throttle) {
			if (isWaking) {
				LOGD("waking worker is going to sleep -> reset to pending");
				auto oldState = resetWaking();
				return oldState == ThrottleState::waking;
			}
		}

		return true;
	}

	auto canWakeup() -> bool {
		if constexpr (emper::WORKER_WAKEUP_STRATEGY == emper::WorkerWakeupStrategy::throttle) {
			return canWakeupThrottle();
		} else {
			return true;
		}
	}

	template <CallerEnvironment callerEnvironment>
	auto getWakeupCount() -> workerid_t {
		if constexpr (emper::WORKER_WAKEUP_STRATEGY == emper::WorkerWakeupStrategy::one) {
			return 1;

		} else if constexpr (emper::WORKER_WAKEUP_STRATEGY == emper::WorkerWakeupStrategy::all) {
			return workerCount;

		} else if constexpr (emper::WORKER_WAKEUP_STRATEGY == emper::WorkerWakeupStrategy::throttle) {
			return getWakeupCountThrottle<callerEnvironment>();

		} else {
			ABORT("Unknown WakeStrategy");
		}
	}
};
}	 // namespace emper
