// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus, Florian Fischer
#include <cstddef>

#include "BinaryPrivateSemaphore.hpp"		 // for BinaryPrivateSemaphore, BPS
#include "CountingPrivateSemaphore.hpp"	 // for CountingPrivateSemaphore, CPS
#include "Fiber.hpp"										 // for Fiber, Fiber::NOT_AFFINE
#include "Runtime.hpp"									 // for Runtime
#include "emper-common.h"								 // for workeraffinity_t
#include "emper-config.h"								 // IWYU pragma: keep
#include "emper.h"											 // for fiber, cps, bps, runtime

#ifdef EMPER_IO
#include <fcntl.h>
#include <sys/socket.h>	 // for socklen_t
#include <sys/types.h>	 // for off_t

#include <cstdarg>	// for va_arg and co. only required if EMPER_IO

#include "io.hpp"	 // for emper::io::emper_*
#endif

auto init_runtime(void) -> runtime* {
	auto* r = new Runtime();
	return reinterpret_cast<runtime*>(r);
}

void wait_until_runtime_finished() {
	Runtime* r = Runtime::getRuntime();
	r->waitUntilFinished();
}

auto fiber_from(void (*function)(void*), void* arg) -> fiber* {
	Fiber* f = Fiber::from(function, arg);
	return reinterpret_cast<fiber*>(f);
}

auto fiber_from0(void (*function)()) -> fiber* {
	Fiber* f = Fiber::from(function);
	return reinterpret_cast<fiber*>(f);
}

auto aff_fiber_from(void (*function)(void*), void* arg, workeraffinity_t* affinity) -> fiber* {
	Fiber* f = Fiber::from(function, arg, affinity);
	return reinterpret_cast<fiber*>(f);
}

auto aff_fiber_from0(void (*function)(), workeraffinity_t* affinity) -> fiber* {
	Fiber* f = Fiber::from(function, affinity);
	return reinterpret_cast<fiber*>(f);
}

void init_affinity(workeraffinity_t affinity[], unsigned int n) {
	for (unsigned int i = 0; i < n; ++i) {
		affinity[i] = Fiber::NOT_AFFINE;
	}
}

void schedule(fiber* fiber) {
	Runtime* runtime = Runtime::getRuntime();
	auto* f = reinterpret_cast<Fiber*>(fiber);
	runtime->schedule(*f);
}

void schedule_from_anywhere(fiber* fiber) {
	Runtime* runtime = Runtime::getRuntime();
	auto* f = reinterpret_cast<Fiber*>(fiber);
	runtime->scheduleFromAnywhere(*f);
}

auto new_binary_sem() -> bps* {
	BPS* sem = new BPS();
	return reinterpret_cast<bps*>(sem);
}

void signal_bps(bps* sem) {
	auto* binaryPrivateSemaphore = reinterpret_cast<BinaryPrivateSemaphore*>(sem);
	binaryPrivateSemaphore->signal();
}

void signal_and_exit_bps(bps* sem) {
	auto* binaryPrivateSemaphore = reinterpret_cast<BinaryPrivateSemaphore*>(sem);
	binaryPrivateSemaphore->signalAndExit();
}

void wait_bps(bps* sem) {
	auto* binaryPrivateSemaphore = reinterpret_cast<BinaryPrivateSemaphore*>(sem);
	binaryPrivateSemaphore->wait();
}

auto new_counting_sem() -> cps* {
	CPS* sem = new CPS();
	return reinterpret_cast<cps*>(sem);
}

auto new_counting_sem_with(unsigned int i) -> cps* {
	CPS* sem = new CPS(i);
	return reinterpret_cast<cps*>(sem);
}

void signal_cps(cps* sem) {
	auto* countingPrivateSemaphore = reinterpret_cast<CountingPrivateSemaphore*>(sem);
	countingPrivateSemaphore->signal();
}

void signal_and_exit_cps(cps* sem) {
	auto* countingPrivateSemaphore = reinterpret_cast<CountingPrivateSemaphore*>(sem);
	countingPrivateSemaphore->signalAndExit();
}

void wait_cps(cps* sem) {
	auto* countingPrivateSemaphore = reinterpret_cast<CountingPrivateSemaphore*>(sem);
	countingPrivateSemaphore->wait();
}

#ifdef EMPER_IO
auto emper_recv(int socket, void* buffer, size_t length, int flags) -> ssize_t {
	return emper::io::recvAndWait(socket, buffer, length, flags);
}

auto emper_send(int socket, const void* buffer, size_t length, int flags) -> ssize_t {
	return emper::io::sendAndWait(socket, buffer, length, flags);
}

auto emper_connect(int socket, const struct sockaddr* address, socklen_t address_len) -> int {
	return emper::io::connectAndWait(socket, address, address_len);
}

auto emper_accept(int socket, struct sockaddr* address, socklen_t* address_len) -> int {
	return emper::io::acceptAndWait(socket, address, address_len);
}

auto emper_read_file(int fildes, void* buf, size_t nbyte, off_t offset) -> ssize_t {
	return emper::io::readFileAndWait(fildes, buf, nbyte, offset);
}

auto emper_write_file(int fildes, const void* buf, size_t nbyte, off_t offset) -> ssize_t {
	return emper::io::writeFileAndWait(fildes, buf, nbyte, offset);
}

auto emper_writev(int fildes, const struct iovec* iov, int iovcnt) -> ssize_t {
	return emper::io::writevAndWait(fildes, iov, iovcnt);
}

// C function overloading inspired by
// http://locklessinc.com/articles/overloading
auto emper_open(const char* pathname, int flags, ...) -> int {
	if (flags | O_CREAT || flags | O_TMPFILE) {
		va_list v;
		va_start(v, flags);

		mode_t mode = va_arg(v, mode_t);
		va_end(v);
		return emper::io::openAndWait(pathname, flags, mode);
	}

	return emper::io::openAndWait(pathname, flags);
}

auto emper_openat(int dirfd, const char* pathname, int flags, ...) -> int {
	if (flags | O_CREAT || flags | O_TMPFILE) {
		va_list v;
		va_start(v, flags);

		mode_t mode = va_arg(v, mode_t);
		va_end(v);
		return emper::io::openatAndWait(dirfd, pathname, flags, mode);
	}

	return emper::io::openatAndWait(dirfd, pathname, flags);
}

auto emper_shutdown(int sockfd, int how) -> int { return emper::io::shutdownAndWait(sockfd, how); }

auto emper_close(int fd) -> int { return emper::io::closeAndWait(fd); }
#endif
