// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#include "Future.hpp"

#include <fcntl.h>
#include <sys/mman.h>
#include <sys/uio.h>
#include <unistd.h>

#include <cassert>	// for assert
#include <cerrno>		// for errno, EAGAIN, EWOULDBLOCK
#include <ctime>
#include <ostream>	// for operator<<, ostream, basic_ost...

#include "CallerEnvironment.hpp"	// for CallerEnvironment
#include "Debug.hpp"							// for LOGD, LOGW
#include "Fiber.hpp"
#include "Runtime.hpp"
#include "Worker.hpp"
#include "io/IoContext.hpp"
#include "io/Operation.hpp"
#include "io/Stats.hpp"
#include "lib/TaggedPtr.hpp"

namespace emper::io {

template <CallerEnvironment callerEnvironment>
auto PartialCompletableFuture::tryComplete(int32_t res)
		-> PartialCompletableFuture::CompletionType {
	// We expect partial completions and made progress
	if (!isDependency() && partialCompletion > DISABLE_PARTIAL_COMPLETION && res > 0) {
		partialCompletion += res;
		// write/send: we have not written the whole buffer yet
		// read: we have not filled the buffer yet
		if ((size_t)partialCompletion < len) {
			IoContext* io = IoContext::getIo<callerEnvironment>();
			io->submit<callerEnvironment>(*this);
			return CompletionType::Resubmission;
		}

		return CompletionType::IncrementalCompletion;
	}

	return CompletionType::Completion;
}

// show the compiler our template incarnations
template PartialCompletableFuture::CompletionType
PartialCompletableFuture::tryComplete<CallerEnvironment::EMPER>(int32_t res);
template PartialCompletableFuture::CompletionType
PartialCompletableFuture::tryComplete<CallerEnvironment::ANYWHERE>(int32_t res);

auto Future::_wait() -> int32_t {
	if constexpr (emper::IO_SYNCHRONOUS) {
		completeSynchronously();
	} else {
		LOGD("Waiting on " << this);

		sem.wait();
	}

	return get();
}

auto Future::waitAndSetErrno() -> int32_t {
	int32_t res = wait();
	// move error codes from io_uring into errno
	if (res < 0) {
		errno = -res;
		return -1;
	}

	return res;
}

void Future::submit() {
	emper::assertInRuntime();
	assert(!isSubmitted());
	IoContext* io = IoContext::getWorkerIo();

	LOGD("submit " << this << " to IoContext " << io);

	io->submitAndReap(*this);
}

auto Future::cancel() -> int32_t {
	LOGD("Cancelling " << this);
	markCancelled();

	if (dependency) {
		dependency->cancel();
	}

	// With the separation of isSubmitted and isPrepared dependent Futures
	// are not marked as submitted anymore.
	// This means we have to check if we are canceling a not submitted Future
	// which is not a dependency.
	if ((!isSubmitted() && !isDependency()) || isForgotten()) {
		LOGW("Cancelling " << (isForgotten() ? "forgotten " : "unsubmitted ") << this);
		return -ENOENT;
	}

	if (isCompleted()) {
		return returnValue;
	}

	// Only try to cancel Futures which have actually reached the io_uring
	if (isPrepared()) {
		CancelWrapper cancellation(*this);
		// We need to cancel on a different worker
		if (!emper::IO_SINGLE_URING && submitter != IoContext::workerIo) {
			submitter->runtime.scheduleOn(*Fiber::from([&cancellation] { cancellation.submit(); }),
																		submitter->worker->getWorkerId());
		} else {
			cancellation.submit();
		}
		const int32_t cancelResult = cancellation.wait();

		if (callback) returnValue = cancelResult;
	}

	// We have to ensure that the Future can be dropped when we return.
	// This is possible when it gets completed or if we are sure it will no longer
	// be use in the IO-subsystem (SubmitActor).
	// Both can be achieved by awaiting it.
	// The former is trivial.
	// The latter is guarantied for Futures that can not be completed
	// (forgotten, callback ones) because we signal their semaphore during
	// preparation and after that they will no longer be used by EMPER.
	// The kernel though may still use resources referenced by the Future such as
	// a buffer or file descriptor.
	return _wait();
}

/**
 * Internal hiding functions for Stats::recordCompletion double dispatch
 *
 * The polymorphism functions are Future::recordCompletion and are defined in
 * Future.hpp so they are included in each compilation unit and the
 * compiler can optimize the recordCompletion calls if stats are off
 */
void Future::recordCompletionInternal(Stats& stats, int32_t res) const {
	stats.recordCompletion(*this, res);
}

void PartialCompletableFuture::recordCompletionInternal(Stats& stats, int32_t res) const {
	stats.recordCompletion(*this, res);
}

void WritevFuture::recordCompletionInternal(Stats& stats, int32_t res) const {
	stats.recordCompletion(*this, res);
}

void SendFuture::submit() {
	if constexpr (emper::IO_TRY_SYSCALL) {
		if (dependency) goto normal_submit;

		ssize_t res = send(fd, buf, len, flags | MSG_DONTWAIT);
		if (res != -1 || (errno != EAGAIN && errno != EWOULDBLOCK)) {
			markSubmitted();
			int ires = res == -1 ? -errno : static_cast<int>(res);
			complete(ires);
			return;
		}
	}

normal_submit:
	Future::submit();
}

void RecvFuture::submit() {
	if constexpr (emper::IO_TRY_SYSCALL) {
		if (dependency) goto normal_submit;

		ssize_t res = recv(fd, buf, len, flags | MSG_DONTWAIT);
		if (res != -1 || (errno != EAGAIN && errno != EWOULDBLOCK)) {
			markSubmitted();
			int ires = res == -1 ? -errno : static_cast<int>(res);
			complete(ires);
			return;
		}
	}

normal_submit:
	Future::submit();
}

void CancelWrapper::prepareSqeInternal(struct io_uring_sqe* sqe) {
	auto& future = *reinterpret_cast<Future*>(buf);
	if constexpr (!emper::IO_SINGLE_URING) {
		assert(future.submitter == IoContext::workerIo);
	}
	uint64_t user_data = IoContext::createFutureTag(future);
	io_uring_prep_cancel64(sqe, user_data, 0);
}

void Future::completeSynchronously() {
	int32_t res = [this]() -> int32_t {
		switch (op) {
			case Operation::SEND:
				return static_cast<int32_t>(::send(fd, buf, len, flags));

			case Operation::RECV:
				return static_cast<int32_t>(::recv(fd, buf, len, flags));

			case Operation::CONNECT:
				return ::connect(fd, (const struct sockaddr*)buf, (socklen_t)len);

			case Operation::ACCEPT:
				// NOLINTNEXTLINE(performance-no-int-to-ptr)
				return ::accept4(fd, (struct sockaddr*)buf, (socklen_t*)len, 0);

			case Operation::OPENAT:
				return ::openat(fd, reinterpret_cast<const char*>(buf), static_cast<int>(len),
												static_cast<mode_t>(offset));

			case Operation::READ:
				if (offset == -1) return static_cast<int32_t>(::read(fd, buf, len));
				return static_cast<int32_t>(::pread(fd, buf, len, offset));

			case Operation::WRITE:
				if (offset == -1) return static_cast<int32_t>(::write(fd, buf, len));
				return static_cast<int32_t>(::pwrite(fd, buf, len, offset));

			case Operation::WRITEV:
				return static_cast<int32_t>(::writev(fd, (const struct iovec*)buf, static_cast<int>(len)));

			case Operation::CLOSE:
				return ::close(fd);

			case Operation::SHUTDOWN:
				return ::shutdown(fd, flags);

			case Operation::LINK_TIMEOUT:
			case Operation::CANCEL:
				DIE_MSG(op << " is not supported synchronously");

			case Operation::TIMEOUT: {
				const auto* fts = (const AlarmFuture::Timespec*)buf;
				struct timespec ts = {.tv_sec = fts->tv_sec, .tv_nsec = fts->tv_nsec};
				::nanosleep(&ts, nullptr);
				return -ETIME;
			}

			case Operation::MADVISE:
				return ::madvise(buf, static_cast<off_t>(len), flags);

			default:
				DIE_MSG("unknown IO operation:" << (int)op);
		}
	}();

	complete(res);
}
}	 // namespace emper::io
