// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#pragma once

#include <liburing.h>		 // for io_uring_prep_read, io_uring_p...
#include <sys/socket.h>	 // for socklen_t
#include <sys/types.h>

#include <atomic>
#include <cassert>
#include <cstddef>	// for size_t
#include <cstdint>	// for int32_t, uint8_t
#include <cstdlib>	// for abort
#include <functional>
#include <ostream>	// for operator<<, ostream, basic_ost...
#include <stdexcept>
#include <string>
#include <utility>

#include "BinaryPrivateSemaphore.hpp"	 // for BPS
#include "BlockablePurpose.hpp"
#include "CallerEnvironment.hpp"
#include "Common.hpp"
#include "Debug.hpp"	// for LOGD, LogSubsystem, LogSubsyst...
#include "Emper.hpp"	// for DEBUG
#include "emper-common.h"
#include "io/Operation.hpp"	 // for Operation, operator<<, Operati...

class Fiber;

namespace emper::io {
class Stats;
class SubmitActor;	// IWYU pragma: keep
class IoContext;

class FutureError : public std::logic_error {
	friend class Future;
	FutureError(const std::string& what) : std::logic_error(what) {}
};

/*
 * @brief Future representing an IO request which can be awaited
 */
class Future : public Logger<LogSubsystem::IO> {
	friend class GlobalIoContext;
	friend class IoContext;
	friend class Stats;
	friend class SubmitActor;
	friend class CancelWrapper;

 public:
	// User facing Callback type
	using Callback = std::function<void(const int32_t&)>;

 protected:
	// Internal Callback type
	class CallbackInternal {
		std::function<void(const int32_t&)> callback;

	 public:
		workeraffinity_t affinity;

		CallbackInternal(Callback callback) : callback(std::move(callback)) {}

		void operator()(const int32_t& res) { callback(res); }
	};

	/*
	 * @brief Semaphore enabling Fibers to wait for the completion of this Future
	 *
	 * The semaphore is either used to block contexts waiting on the future
	 * or used to ensure that the future is not dropped and thus possibly deleted
	 * until it was prepared by the submit actor (only in the IO_SINGLE_URING case).
	 */
	BPS sem;

	using State = struct State {
		enum _State {
			Submitted = 1 << 0,	 /*!< This Future was submitted to the IO subsystem */
			Prepared = 1 << 1,	 /*!< An sqe for this Future was prepared */
			Completed = 1 << 2,	 /*!< The semaphor was signaled  */
			Retrieved = 1 << 3,	 /*!< Future::wait() has retrieved the result */
			Dependency = 1 << 4, /*!< This Future is a dependency of another */
			Cancelled = 1 << 5,	 /*!< Future::cancel() was called */
			Forgotten = 1 << 6,	 /*!< This Future will not be awaited */
		};
	};

	/**
	 * @brief State of the Future object
	 *
	 * This state member is used to track the state of the future and determines
	 * our use of it.
	 * Most of the time the state is not use concurrently but some data races exist:
	 * - cancel() & complete(): They race for state.completed when complete() may
	 *                          run in a different thread than the call to submit().
	 *                          This can happen when a completer is present, or we use a
	 *                          single io_uring or IO-stealing is enabled.
	 *                          This race is **NOT HARMFULL** since when cancel
	 *                          wins the race the only thing it does is calling wait().
	 *
	 * - cancel() & prepare():  They race for state.prepared when prepareSqe() may
	 *                          run in a different thread than cancel().
	 *                          This can happen when we use a single io_uring.
	 *                          The race is **NOT HARMFULL** we always await the
	 *                          future's semaphore in cancel. If the Future is
	 *                          not completeable the semaphore will be signaled after
	 *                          preparation otherwise when completed.
	 */
	uint8_t state = 0;

	/**
	 * @brief atomic helper to set a bit in the state in the face of data races
	 */
	void casIntoState(uint8_t bit) {
		auto* astate = reinterpret_cast<std::atomic<uint8_t>*>(&state);
		uint8_t oldState = astate->load(std::memory_order_relaxed);
		do {
		} while (!astate->compare_exchange_weak(oldState, oldState | bit, std::memory_order_release,
																						std::memory_order_relaxed));
	}

	/**
	 * @brief atomic helper to acquire the state
	 */
	auto acquireState() const -> uint8_t {
		auto* astate = reinterpret_cast<const std::atomic<uint8_t>*>(&state);
		return astate->load(std::memory_order_acquire);
	}

	IoContext* submitter = nullptr;

	/* IO operation to perform */
	const Operation op;

	/* Arguments for the IO operation */
	const int fd;
	void* const buf;
	const size_t len;
	union {
		const off_t offset;
		const int flags;
	};

	/**
	 * Return value of the operation.
	 * __s32 is used by liburing for the res field in its cqe struct
	 * This value is returned by Future::wait() and can only differentiate between
	 * failure or full success.
	 */
	int32_t returnValue = 0;

	// Dependency futures will be submitted to the io_uring before
	// this and will be linked to this IO request using IOSQE_IO_LINK
	Future* dependency = nullptr;

	/**
	 * Callback function which is called by a new scheduled Fiber on completion
	 * A new Callback object is heap allocated in setCallback and freed by the
	 * executing Fiber after callback returned;
	 * The heap allocation and thus the Callback pointer is required to decouple the lifetimes
	 * of the Future used for submission and the Callback which must life for the
	 * time the request is in the io_uring.
	 * The callback is called with the value causing the completion
	 */
	CallbackInternal* callback = nullptr;

	void prepareSqe(IoContext& io, struct io_uring_sqe* sqe) {
		submitter = &io;
		markPrepared();
		prepareSqeInternal(sqe);
	}

	virtual void prepareSqeInternal(struct io_uring_sqe* sqe) = 0;

	void submitAndForget() {
		markForgotten();
		submit();
		// When using a SubmitActor a Future may be submitted to the SubmitActor but
		// no SQE was prepared for it yet.
		// It must be ensured that the Future is prepared before forgetting and possibly dropping it.
		// We await its preperation using the nomal BPS which is not used otherwise for forgotten
		// Futures.
		// The BPS of forgotten Futures is signaled in IoContext::prepareFutureChain
		// when using a single io_uring
		if constexpr (emper::IO_SINGLE_URING) {
			sem.wait();
		}
	}

	void setCompletion(int32_t res) {
		assert(!isCompleted());
		LOGD("Complete " << this << " with result " << res);
		returnValue = res;
		markCompleted();
	}

	virtual void complete(int32_t res) {
		setCompletion(res);
		sem.signal();
	}

	virtual void completeFromAnywhere(int32_t res) {
		setCompletion(res);
		sem.signalFromAnywhere();
	}

	virtual auto completeAndGetContinuation(int32_t res) -> Fiber* {
		setCompletion(res);
		return sem.signalAndGetContinuation();
	}

	/**
	 * Used for Stats::recordCompletion double dispatch
	 */
	virtual void recordCompletion(Stats& stats, int32_t res) {
		if constexpr (emper::STATS) {
			recordCompletionInternal(stats, res);
		}
	}

	void recordCompletionInternal(Stats& stats, int32_t res) const;

	void completeSynchronously();

	/**
	 * @brief Block till the IO request is completed without sanity checks
	 *
	 * @return return the result received from the io_uring
	 *
	 * It is guarantied that after wait() returned the memory of a Future and
	 * all necessary resources for the operation are no longer used and
	 * referenced by EMPER and thus can be freed.
	 */
	auto _wait() -> int32_t;

	Future(Operation op, int fd, void* buf, size_t len, off_t offset)
			: sem(emper::BlockablePurpose::IO), op(op), fd(fd), buf(buf), len(len), offset(offset) {};

	Future(Operation op, int fd, void* buf, size_t len, int flags)
			: sem(emper::BlockablePurpose::IO), op(op), fd(fd), buf(buf), len(len), flags(flags) {};

 public:
	// Clang-tidy warns about the exception possibly thrown by
	// ~Future -> cancel -> wait
	// But this exception will never be thrown because wait() throws an exception
	// only if callback is set and if callback is set ~Future does not call cancel.
	// NOLINTNEXTLINE(bugprone-exception-escape)
	virtual ~Future() {
		if (isForgotten() || isRetrieved() || isCancelled()) {
			return;
		}

		// If this Future uses a callback we have to ensure that the future is
		// prepared before dropping it.
		if (callback && isPrepared()) return;

		if constexpr (emper::DEBUG) {
			if (!isDependency() && !callback) {
				LOGE(this << " created but never awaited");
				abort();
			}
		}

		// Cancel this Future to guaranty memory safety.
		// By canceling we assure that the this pointer and the contained buffer
		// passed to io_uring is not used any more by the kernel or the IoContext.
		cancel();
	}

	// State setters
	// Those are called during the life of a Future and may introduce data races.
	void markSubmitted() { state |= State::Submitted; }
	void markPrepared() {
		if constexpr (emper::IO_SINGLE_URING) {
			casIntoState(State::Prepared);
		} else {
			state |= State::Prepared;
		}
	}
	void markCompleted() { state |= State::Completed; }
	void markRetrieved() { state |= State::Retrieved; }
	void markCancelled() {
		if constexpr (emper::IO_SINGLE_URING) {
			casIntoState(State::Cancelled);
		} else {
			state |= State::Cancelled;
		}
	}

	// Those are state information which can not change after submitting a Future.
	void markAsDependency() {
		assert(!isSubmitted());
		state |= State::Dependency;
	}
	void markForgotten() {
		assert(!isSubmitted());
		state |= State::Forgotten;
	}

	// State getters
	[[nodiscard]] auto isSubmitted() const -> bool { return state & State::Submitted; }

	[[nodiscard]] auto isPrepared() const -> bool {
		if constexpr (emper::IO_SINGLE_URING) {
			return acquireState() & State::Prepared;
		}
		return state & State::Prepared;
	}

	[[nodiscard]] auto isCompleted() const -> bool { return state & State::Completed; }
	[[nodiscard]] auto isRetrieved() const -> bool { return state & State::Retrieved; }
	[[nodiscard]] auto isDependency() const -> bool { return state & State::Dependency; }

	[[nodiscard]] auto isCancelled() const -> bool {
		if constexpr (emper::IO_SINGLE_URING) {
			return acquireState() & State::Cancelled;
		}
		return state & State::Cancelled;
	}

	[[nodiscard]] auto isForgotten() const -> bool { return state & State::Forgotten; }

	/*
	 * @brief reset the Future
	 *
	 * Reset the state and the BPS.
	 * Resetting a Future is useful if we wan't to use the same Future object
	 * multiple times.
	 * A Future for a specific reoccurring IO operation can be created and reused
	 * in a submit, reset loop.
	 */
	virtual inline void reset() {
		LOGD("Resetting Future");
		state = 0;
		sem.reset();
	}

	/*
	 * @brief set Future as dependency
	 *
	 * The dependency must complete before this future.
	 * This is assured by linking the futures' sqes in the io_uring using the IOSQE_IO_LINK flag.
	 * If a request in the link fails all dependent futures will return -ECANCELED.
	 *
	 * See: <a href="https://unixism.net/loti/tutorial/link_liburing.html">Liburing linking
	 * requests</a>
	 * The emper equivalent to the example from the liburing documentation.
	 *
	 * @code
	 * int fd = open(FILE_NAME, O_RDWR|O_TRUNC|O_CREAT, 0644);
	 * if (fd < 0 ) {
	 * perror("open");
	 * return 1;
	 * }
	 *
	 * char* msg = "Hallo emper!\n";
	 * WriteFuture writeFuture(fd, &msg, strlen(msg), 0);
	 *
	 * char buf[32];
	 * ReadFuture readFuture(fd, &buf, sizeof(buf), 0);
	 * readFuture.setDependency(writeFuture);
	 *
	 * CloseFuture closeFuture(fd);
	 * closeFuture.setDependency(readFuture);
	 *
	 * // Blocks till all three futures are completed
	 * int bytes_read = closeFuture.submitAndWait();
	 *
	 * // So request in the chain could not be fully completed
	 * if (bytes_read == -ECANCELED) {
	 * }
	 * @endcode
	 *
	 */
	inline void setDependency(Future& dependency) {
		dependency.markAsDependency();
		this->dependency = &dependency;
	}

	/*
	 * @brief register a callback which is executed in a new Fiber on completion
	 *
	 * @param callback Callback reference which is copied and executed on completion.
	 *                 It gets passed the value causing the completion.
	 */
	inline void setCallback(const Callback& callback) {
		if (unlikely(this->callback)) {
			delete this->callback;
		}
		this->callback = new CallbackInternal(callback);
	}

	/*
	 * @brief submit Future for asynchronous completion to an IoContext
	 */
	virtual void submit();

	/**
	 * @brief cancel Future and wait if necessary for its cancellation
	 *
	 * @return return the result received from the io_uring
	 *
	 * It is guarantied that after cancel() returned the memory of a Future and
	 * all necessary resources for the operation are no longer used and
	 * referenced by EMPER and thus can be freed.
	 */
	auto cancel() -> int32_t;

	/**
	 * @brief Block till the IO request is completed
	 *
	 * @return return the result received from the io_uring
	 *
	 * It is guarantied that after wait() returned the memory of a Future and
	 * all necessary resources for the operation are no longer used and
	 * referenced by EMPER and thus can be freed.
	 */
	auto wait() -> int32_t {
		if (unlikely(callback)) {
			throw FutureError("Futures with registered callback must not be awaited");
		}

		return _wait();
	}

	/**
	 * @brief Get the return value of the Future
	 *
	 * @return return the result received from the io_uring
	 *
	 * Only call get() if you have ensured that the future was completed.
	 * By calling either wait() or cancel() beforehand.
	 * Get can be used to check the return value of Futures canceled as part of
	 * a Future chain.
	 */
	auto get() -> int32_t {
		LOGD("get return value of " << this);

		if constexpr (emper::DEBUG) {
			markRetrieved();
		}

		return returnValue;
	}

	/**
	 * @brief Block till the IO request is completed and set errno on error
	 *
	 * @return -1 on error and errno will be set, otherwise the return value of the IO request
	 */
	auto waitAndSetErrno() -> int32_t;

	/**
	 * @brief Equivalent to calling wait() after calling submit()
	 */
	inline auto submitAndWait() -> int32_t {
		submit();
		return wait();
	}

	friend auto operator<<(std::ostream& os, const Future& f) -> std::ostream& {
		return os << f.op << " Future " << (void*)&f;
	}

	friend auto operator<<(std::ostream& os, const Future* f) -> std::ostream& {
		if (f) {
			return os << *f;
		}

		return os << "Future nullptr";
	}
};

class PartialCompletableFuture : public Future {
	friend class Stats;
	friend class IoContext;
	friend class RecvFuture;

	enum class CompletionType {
		Resubmission,
		IncrementalCompletion,
		Completion,
	};

	/**
	 * @brief Try to complete this IO request using the result of an IO operation.
	 *        If the IO request is completed return true otherwise resubmit it.
	 *
	 * @param res Result from an IO operation from a io_uring_cqe or a common syscall
	 *
	 * @return the type of completion
	 */
	template <CallerEnvironment callerEnvironment>
	auto tryComplete(int32_t res) -> CompletionType;

 protected:
	/**
	 * This member is used internally to handle partial completion.
	 * If initialized with -1 it signals that the IO request should immediately return to the user.
	 * A value >= 0 signals that this IO request should not * return to the user until it is
	 * fully completed or a persistent error arises.
	 * The IO request is considered fully completed if partialCompletion == len.
	 *
	 * When the IO request returns to the user the combination of @c future.returnValue and
	 * @c future.partialCompletion can signal partial completion.
	 */
	int32_t partialCompletion;

	// return immediately to the user
	static const int32_t DISABLE_PARTIAL_COMPLETION = -1;
	// try to fully complete the request
	static const int32_t ENABLE_PARTIAL_COMPLETION = 0;

	PartialCompletableFuture(Operation op, int fd, void* buf, size_t len, off_t offset,
													 bool completeFully)
			: Future(op, fd, buf, len, offset),
				partialCompletion(completeFully ? ENABLE_PARTIAL_COMPLETION : DISABLE_PARTIAL_COMPLETION) {
				};

	PartialCompletableFuture(Operation op, int fd, void* buf, size_t len, int flags,
													 bool completeFully)
			: Future(op, fd, buf, len, flags),
				partialCompletion(completeFully ? ENABLE_PARTIAL_COMPLETION : DISABLE_PARTIAL_COMPLETION) {
				};

	PartialCompletableFuture(Operation op, int fd, void* buf, size_t len, off_t offset,
													 int32_t partialCompletion)
			: Future(op, fd, buf, len, offset), partialCompletion(partialCompletion) {};

	PartialCompletableFuture(Operation op, int fd, void* buf, size_t len, int flags,
													 int32_t partialCompletion)
			: Future(op, fd, buf, len, flags), partialCompletion(partialCompletion) {};

	/**
	 * Used for Stats::recordCompletion double dispatch
	 */
	void recordCompletion(Stats& stats, int32_t res) override {
		if constexpr (emper::STATS) {
			recordCompletionInternal(stats, res);
		}
	}

	void recordCompletionInternal(Stats& stats, int32_t res) const;

	void complete(int32_t res) override {
		CompletionType completion = tryComplete<CallerEnvironment::EMPER>(res);

		switch (completion) {
			case CompletionType::Resubmission:
				return;
			case CompletionType::IncrementalCompletion:
				Future::complete(partialCompletion);
				return;
			case CompletionType::Completion:
				Future::complete(res);
				return;
		}
	}

	void completeFromAnywhere(int32_t res) override {
		CompletionType completion = tryComplete<CallerEnvironment::ANYWHERE>(res);

		switch (completion) {
			case CompletionType::Resubmission:
				return;
			case CompletionType::IncrementalCompletion:
				Future::completeFromAnywhere(partialCompletion);
				return;
			case CompletionType::Completion:
				Future::completeFromAnywhere(res);
				return;
		}
	}

	auto completeAndGetContinuation(int32_t res) -> Fiber* override {
		CompletionType completion;

		// If there is no completer we must be in a emper worker
		// Otherwise decide at runtime where the possible partialCompletion should be resubmitted
		if (emper::IO_COMPLETER_BEHAVIOR == emper::IoCompleterBehavior::none || emper::inRuntime()) {
			completion = tryComplete<CallerEnvironment::EMPER>(res);
		} else {
			completion = tryComplete<CallerEnvironment::ANYWHERE>(res);
		}

		switch (completion) {
			case CompletionType::Resubmission:
				return nullptr;
			case CompletionType::IncrementalCompletion:
				return Future::completeAndGetContinuation(partialCompletion);
			case CompletionType::Completion:
				return Future::completeAndGetContinuation(res);
			default:
				DIE_MSG("Unknown CompletionType: " << (int)completion);
		}
	}

 public:
	/**
	 * @brief Block till the IO request is completed
	 *
	 * @param[out] partialCompletion The amount of work done till an persistent error arose
	 *
	 * @return return the result received from the io_uring
	 */
	auto wait(int32_t& partialCompletion) -> int32_t {
		int32_t res = Future::wait();
		partialCompletion = this->partialCompletion;
		return res;
	}
	using Future::wait;

	/*
	 * @brief Equivalent to calling wait(partialCompletion) after calling submit()
	 */
	inline auto submitAndWait(int32_t& partialCompletion) -> int32_t {
		Future::submit();
		return wait(partialCompletion);
	}
	using Future::submitAndWait;

	/**
	 * @brief reset the Future and its partialCompletion.
	 */
	void reset() override {
		if (partialCompletion != DISABLE_PARTIAL_COMPLETION) {
			partialCompletion = ENABLE_PARTIAL_COMPLETION;
		}

		Future::reset();
	}
};

class SendFuture : public PartialCompletableFuture {
	void prepareSqeInternal(struct io_uring_sqe* sqe) override {
		if (partialCompletion == DISABLE_PARTIAL_COMPLETION) {
			io_uring_prep_send(sqe, fd, buf, len, flags);
		} else {
			io_uring_prep_send(sqe, fd, (char*)buf + partialCompletion, len - partialCompletion, flags);
		}
	}

 public:
	SendFuture(int socket, const void* buffer, size_t length, int flags, int32_t partialCompletion)
			: PartialCompletableFuture(Operation::SEND, socket, const_cast<void*>(buffer), length, flags,
																 partialCompletion) {};

	SendFuture(int socket, const void* buffer, size_t length, int flags, bool send_all = true)
			: SendFuture(socket, const_cast<void*>(buffer), length, flags,
									 send_all ? ENABLE_PARTIAL_COMPLETION : DISABLE_PARTIAL_COMPLETION) {};

	void submit() override;
};

class RecvFuture : public Future {
	void prepareSqeInternal(struct io_uring_sqe* sqe) override {
		io_uring_prep_recv(sqe, fd, buf, len, flags);
	}

 public:
	RecvFuture(int socket, void* buffer, size_t length, int flags)
			: Future(Operation::RECV, socket, buffer, length, flags) {};

	void submit() override;
};

class ConnectFuture : public Future {
	void prepareSqeInternal(struct io_uring_sqe* sqe) override {
		io_uring_prep_connect(sqe, fd, (const struct sockaddr*)buf, (socklen_t)len);
	}

 public:
	ConnectFuture(int socket, const struct sockaddr* address, socklen_t address_len)
			: Future(Operation::CONNECT, socket, (void*)address, address_len, 0) {};
};

class AcceptFuture : public Future {
	void prepareSqeInternal(struct io_uring_sqe* sqe) override {
		// NOLINTNEXTLINE(performance-no-int-to-ptr)
		io_uring_prep_accept(sqe, fd, (struct sockaddr*)buf, (socklen_t*)len, 0);
	}

 public:
	AcceptFuture(int socket, const struct sockaddr* address, socklen_t* address_len)
			: Future(Operation::ACCEPT, socket, (void*)address, (size_t)address_len, 0) {};
};

class ReadFuture : public PartialCompletableFuture {
	void prepareSqeInternal(struct io_uring_sqe* sqe) override {
		if (partialCompletion == DISABLE_PARTIAL_COMPLETION) {
			io_uring_prep_read(sqe, fd, buf, len, offset);
		} else {
			io_uring_prep_read(sqe, fd, (char*)buf + partialCompletion, len - partialCompletion,
												 offset + partialCompletion);
		}
	}

 public:
	ReadFuture(int fildes, void* buf, size_t nbyte, off_t offset, bool read_all = false)
			: PartialCompletableFuture(
						Operation::READ, fildes, buf, nbyte, offset,
						read_all ? ENABLE_PARTIAL_COMPLETION : DISABLE_PARTIAL_COMPLETION) {};
};

class OpenatFuture : public Future {
	void prepareSqeInternal(struct io_uring_sqe* sqe) override {
		io_uring_prep_openat(sqe, fd, reinterpret_cast<const char*>(buf), static_cast<int>(len),
												 static_cast<mode_t>(offset));
	}

 public:
	OpenatFuture(int dirfd, const void* pathname, int flags, mode_t mode = 0)
			: Future(Operation::OPENAT, dirfd, const_cast<void*>(pathname), flags,
							 static_cast<off_t>(mode)) {};
};

class WriteFuture : public PartialCompletableFuture {
	void prepareSqeInternal(struct io_uring_sqe* sqe) override {
		if (partialCompletion == DISABLE_PARTIAL_COMPLETION) {
			io_uring_prep_write(sqe, fd, buf, len, offset);
		} else {
			// TODO: think about partial writes with incremental offset!
			// Are transparent incremental writes on files without O_APPEND even reasonable?
			io_uring_prep_write(sqe, fd, (char*)buf + partialCompletion, len - partialCompletion,
													offset + partialCompletion);
		}
	}

 public:
	WriteFuture(int fildes, const void* buf, size_t nbyte, off_t offset, bool write_all = true)
			: PartialCompletableFuture(
						Operation::WRITE, fildes, const_cast<void*>(buf), nbyte, offset,
						write_all ? ENABLE_PARTIAL_COMPLETION : DISABLE_PARTIAL_COMPLETION) {};
};

class WritevFuture : public Future {
	void prepareSqeInternal(struct io_uring_sqe* sqe) override {
		io_uring_prep_writev(sqe, fd, (const struct iovec*)buf, len, 0);
	}

	/**
	 * Used for Stats::recordCompletion double dispatch
	 */
	void recordCompletion(Stats& stats, int32_t res) override {
		if constexpr (emper::STATS) {
			recordCompletionInternal(stats, res);
		}
	}

	void recordCompletionInternal(Stats& stats, int32_t res) const;

 public:
	WritevFuture(int fildes, const struct iovec* iov, int iovnct)
			: Future(Operation::WRITEV, fildes, (void*)iov, iovnct, 0) {};
};

class CloseFuture : public Future {
	void prepareSqeInternal(struct io_uring_sqe* sqe) override { io_uring_prep_close(sqe, fd); }

 public:
	CloseFuture(int fildes) : Future(Operation::CLOSE, fildes, nullptr, 0, 0) {};

	void submitAndForget() { Future::submitAndForget(); }
};

class ShutdownFuture : public Future {
	void prepareSqeInternal(struct io_uring_sqe* sqe) override {
		io_uring_prep_shutdown(sqe, fd, flags);
	}

 public:
	ShutdownFuture(int sockfd, int how) : Future(Operation::SHUTDOWN, sockfd, nullptr, 0, how) {};
};

/*
 * @brief Add a timeout to any Future not already submitted.
 *
 * A Timeout is added to a Future by inserting an dependent IOURING_OP_LINK_TIMEOUT
 * sqe after the actual IO operation.
 * The dependent timeout will not be started like usual dependent requests after
 * the previous request is completed, it is instead armed immediately when the
 * actual request is started.
 * See: https://lwn.net/Articles/803932/
 *
 * @code
 * ReadFuture rf(fd, buf, buf_len, 0);
 * struct __kernel_timespec ts = {.tv_sec = 0, .tv_nsec = 0);
 * TimeoutWrapper t(rf, ts);
 * // Both futures should be awaited by calling its wait method to
 * // prevent use after free problems.
 * t.submitAndWait()
 * rf.Wait()
 * @endcode
 */
class TimeoutWrapper : public Future {
 public:
	using Timespec = struct __kernel_timespec;

	TimeoutWrapper(Future& future, Timespec& ts)
			: Future(Operation::LINK_TIMEOUT, 0, (void*)&ts, 0, 0) {
		setDependency(future);
	};

 private:
	void prepareSqeInternal(struct io_uring_sqe* sqe) override {
		io_uring_prep_link_timeout(sqe, (Timespec*)buf, 0);
	}
};

/*
 * @brief Arm a timeout which will signal the future when it is reached
 *
 * If the timeout was reached calling wait() will return -ETIME.
 */
class AlarmFuture : public Future {
 public:
	using Timespec = struct __kernel_timespec;

	AlarmFuture(Timespec& ts) : Future(Operation::TIMEOUT, 0, (void*)&ts, 0, 0) {};

 private:
	void prepareSqeInternal(struct io_uring_sqe* sqe) override {
		io_uring_prep_timeout(sqe, (Timespec*)buf, 0, 0);
	}
};

/*
 * @brief Request the cancellation of another Future and await the cancellation
 *
 * If the target was found and cancelled wait() will immediately return 0.
 * If the target was already completed wait() will immediately return -ENOENT.
 * If the target was already started wait() will return -EALREADY.
 */
class CancelWrapper : public Future {
	void prepareSqeInternal(struct io_uring_sqe* sqe) override;

 public:
	CancelWrapper(Future& future) : Future(Operation::CANCEL, 0, &future, 0, 0) {};
};

/*
 * @brief Request a madvise operation
 */
class MadviseFuture : public Future {
	void prepareSqeInternal(struct io_uring_sqe* sqe) override {
		// TODO: Check that len is within the bounds of off_t.
		auto len_off_t = static_cast<off_t>(len);
		io_uring_prep_madvise(sqe, buf, len_off_t, flags);
	}

 public:
	MadviseFuture(void* addr, size_t len, int advise)
			: Future(Operation::MADVISE, 0, addr, len, advise) {};

	void submitAndForget() { Future::submitAndForget(); }
};
}	 // namespace emper::io
