// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021 Florian Fischer
#include "io/SubmitActor.hpp"

#include <liburing.h>

#include <atomic>		// for atomic
#include <cassert>	// for assert
#include <cerrno>		// for EBUSY, ECANCELED, errno
#include <ostream>	// for operator<<, ostream
#include <vector>		// for vector

#include "CallerEnvironment.hpp"
#include "Common.hpp"	 // for unlikely, DIE_MSG_ERRNO, TIME_NS
#include "Emper.hpp"	 // for DEBUG
#include "Runtime.hpp"
#include "io/Future.hpp"
#include "io/IoContext.hpp"

namespace emper::io {
void SubmitActor::receive(Future* future) {
	LOGD("SubmitActor received " << future);
	assert(future);
	assert(future->isSubmitted());

	// If someone already canceled the future before the submitter
	// receives the future complete the chain with -ECANCELED and return
	if (future->isCancelled()) {
		// complete the chain
		for (Future* f = future; f; f = f->dependency) {
			LOGD("SubmitActor cancel " << f);
			f->recordCompletion(io.stats, -ECANCELED);
			f->complete(-ECANCELED);
		}
		return;
	}

	LOGD("submitting " << future << (future->dependency ? " and it's dependencies" : ""));
	unsigned prepared = io.prepareFutureChain(*future);

	// submit the Future to the io_uring
	int submitted = io_uring_submit(&io.ring);

	// We can't submit our sqe because the CQ is full -> yield
	if (unlikely(submitted == -EBUSY)) {
		if constexpr (emper::DEBUG) {
			LOGGER_LOGI("io_uring_submit returned EBUSY trying to submit in addition to "
									<< io.reqs_in_uring);
		} else {
			LOGW("io_uring_submit returned EBUSY");
		}

		TIME_NS(
				{
					do {
						runtime.yield();
					} while ((submitted = io_uring_submit(&io.ring)) == -EBUSY);
				},
				io.stats.record_io_submit_full_cq);
	}

	if (unlikely(submitted < 0)) {
		errno = -submitted;
		DIE_MSG_ERRNO("io_uring_submit failed");
	}

	// We submitted some Futures to the io_uring
	// Because we submit every Future right away multiple prepared sqes can only
	// occur for Future chains.
	// If we could not submit the whole chain cancel all non submitted Futures
	// because we can not guaranty the soundness of the chain
	// req1 -> invalid_req -> req3
	// will submit only 2 instead of all 3 prepared sqes
	// See: https://github.com/axboe/liburing/issues/186
	if (unlikely(static_cast<unsigned>(submitted) < prepared) &&
			emper::IO_MUST_INVALIDATE_BROKEN_CHAIN) {
		io.cancelUnsubmittedChainParts<CallerEnvironment::EMPER>(*future);
	}

	io.preparedSqes.clear();
}
}	 // namespace emper::io
