// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021-2022 Florian Fischer
#include "lib/MappedFileBuffer.hpp"

#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <cstring>
#include <ostream>

#include "Common.hpp"

namespace emper::lib {

MappedFileBuffer::MappedFileBuffer(const std::string& file) : file(file) {
	fd = open(file.c_str(), O_RDWR | O_TRUNC | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (fd == -1) {
		DIE_MSG_ERRNO("opening file " << file << " failed");
	}

	for (size_t i = 0; i < BUFFER_COUNT; ++i) {
		const auto off = static_cast<off_t>(i * BUFFER_SIZE);
		void* res = mmap(nullptr, BUFFER_SIZE, PROT_WRITE, MAP_FILE | MAP_SHARED, fd, off);
		// NOLINTNEXTLINE(performance-no-int-to-ptr)
		if (res == MAP_FAILED) {
			DIE_MSG_ERRNO("mmap failed");
		}
		bufs[i] = static_cast<char*>(res);
	}

	if (ftruncate(fd, BUFFER_COUNT * BUFFER_SIZE)) {
		DIE_MSG_ERRNO("initial mapped file " << file << "truncation failed");
	}
}

void MappedFileBuffer::trim() {
	const auto finalPos = static_cast<off_t>(bufPos.load(std::memory_order_relaxed));
	if (ftruncate(fd, finalPos)) {
		DIE_MSG_ERRNO("trimming mapped file " << file << " failed");
	}
}

MappedFileBuffer::~MappedFileBuffer() {
	trim();

	for (auto* buf : bufs) {
		if (munmap(buf, BUFFER_SIZE)) {
			DIE_MSG_ERRNO("munmap failed");
		}
	}

	close(fd);
}

void MappedFileBuffer::put(const std::string& message) {
	const size_t messageLen = message.size();
	const size_t startPos = bufPos.fetch_add(messageLen, std::memory_order_relaxed);
	const size_t endPos = startPos + messageLen - 1;
	char* buf = getBuf(startPos);
	const size_t bufOffset = startPos % BUFFER_SIZE;
	char* msgEndBuf = getBuf(endPos);

	// The message fits completly into the current buffer
	if (buf == msgEndBuf) {
		// NOLINTNEXTLINE(bugprone-not-null-terminated-result)
		memcpy(&buf[bufOffset], message.c_str(), message.size());

		// Are we still in the active buffer?
		if (bufOffset > 0) {
			return;
		}
		// The message spans the active and the next buffer
	} else {
		const size_t leftInActive = BUFFER_SIZE - (startPos % BUFFER_SIZE);
		// NOLINTNEXTLINE(bugprone-not-null-terminated-result)
		memcpy(&buf[bufOffset], message.c_str(), leftInActive);
		// NOLINTNEXTLINE(bugprone-not-null-terminated-result)
		memcpy(&msgEndBuf[0], &message.c_str()[leftInActive], message.size() - leftInActive);
	}

	// This is the first overflow there is no old buffer to remap
	if (bufPos < BUFFER_SIZE * 2) {
		return;
	}

	// We are the first writing to the next buffer therefore we are responsible
	// to remap the old.

	// NOTE: This is NOT sound it is possible that other threads are still using the old
	// buffer. But with big enough BUFFER_SIZE it is likely that all threads are using
	// the current active buffer

	// Active buffer -> old buffer
	// Fresh buffer -> active buffer
	// Remap old buffer -> next fresh buffer

	// Remap the old buffer
	const size_t oldBufIdx = ((startPos / BUFFER_SIZE) - 1) % BUFFER_COUNT;
	char* oldBufPtr = bufs[oldBufIdx];

	// The buffer [0, BUFFER_SIZE) is the first active buffer
	// Therefore at bufPos = BUFFER_SIZE + 1 we are in the second active buffer
	const size_t nthActive = (endPos / BUFFER_SIZE) + 1;

	// Our mapped file has the size of all ever active buffers plus the new fresh one
	const auto fileSize = static_cast<off_t>((nthActive + 1) * BUFFER_SIZE);

	// Grow the mapped file
	if (ftruncate(fd, fileSize)) {
		DIE_MSG_ERRNO("growing mapped file " << file << " failed");
	}

	const auto nextOffset = static_cast<off_t>(nthActive * BUFFER_SIZE);
	void* res = mmap(nullptr, BUFFER_SIZE, PROT_WRITE, MAP_FILE | MAP_SHARED, fd, nextOffset);
	// NOLINTNEXTLINE(performance-no-int-to-ptr)
	if (res == MAP_FAILED) {
		DIE_MSG_ERRNO("mmap of fresh buffer failed");
	}
	bufs[oldBufIdx] = static_cast<char*>(res);

	if (munmap(oldBufPtr, BUFFER_SIZE)) {
		DIE_MSG_ERRNO("munmap of replaced buffer failed");
	}
}

}	 // namespace emper::lib
