// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Schmaus, Florian Fischer
#pragma once

#include <cassert>
#include <cstdint>
#include <iosfwd>
#include <string>
#include <type_traits>

namespace emper::lib::math {

// Template magic, thanks to Tobias Langer for help with this.
template <typename Type, typename Enable = void>
struct make_signed_cond;

template <typename Type>
struct make_signed_cond<Type, typename std::enable_if<std::is_floating_point<Type>::value>::type> {
	using type = Type;
};

template <typename Type>
struct make_signed_cond<Type, typename std::enable_if<!std::is_floating_point<Type>::value>::type> {
	using type = typename std::make_signed<Type>::type;
};

// Source of the algorithm:
// https://math.stackexchange.com/questions/106700/incremental-averageing
template <typename average_type, typename counter_type, typename value_type>
void calcRunningAvg(average_type& runningAverage, counter_type count, value_type value) {
	using signed_average_type = typename make_signed_cond<average_type>::type;
	using signed_value_type = typename make_signed_cond<value_type>::type;

	signed_average_type diff =
			((signed_value_type)value - (signed_average_type)runningAverage) / (signed_average_type)count;

	assert(diff >= 0 || (value_type)-diff <= (value_type)runningAverage);

	runningAverage += diff;
}

template <typename average_type = double, typename counter_type = uint32_t>
class RunningAverage {
 private:
	counter_type counter;
	average_type average;

 public:
	RunningAverage() : counter(0), average(0) {};

	template <typename value_type>
	void update(value_type value) {
		counter++;

		calcRunningAvg(average, counter, value);
	}

	[[nodiscard]] auto getCounter() const -> counter_type { return counter; }

	[[nodiscard]] auto getAverage() const -> average_type { return average; }

	friend auto operator<<(std::ostream& strm, const RunningAverage& avg) -> std::ostream& {
		strm << std::to_string(avg.average) << " [" << std::to_string(avg.counter) << "]";
		return strm;
	}
};

constexpr size_t roundUp(size_t number, size_t multiple) {
	size_t multiple_count = (number + multiple - 1) / multiple;
	return multiple_count * multiple;
}
}	 // namespace emper::lib::math
