// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021 Florian Fischer
#include "lib/sync/SpuriousFutex2Semaphore.hpp"

#include <linux/futex.h>
#include <unistd.h>

#include <atomic>
#include <cassert>
#include <cerrno>

#include "Worker.hpp"
#include "emper-config.h"	 // IWYU pragma: keep

#ifndef SYS_futex_waitv
#define SYS_futex_waitv 449
#endif

#ifndef FUTEX_32
#define FUTEX_32 2
#endif

#ifdef EMPER_DEFINE_FUTEX_WAITV
struct futex_waitv {
	uint64_t val;
	uint64_t uaddr;
	uint32_t flags;
	// NOLINTNEXTLINE(bugprone-reserved-identifier)
	uint32_t __reserved;
};
#endif

static void init_futex_waitv(struct futex_waitv* waiter, uint64_t val, void* uaddr) {
	waiter->val = val;
	waiter->uaddr = reinterpret_cast<uintptr_t>(uaddr);
	waiter->flags = FUTEX_PRIVATE_FLAG | FUTEX_32;
	waiter->__reserved = 0;
}

namespace emper::lib::sync {

void SpuriousFutex2Semaphore::wait() {
	const workerid_t workerId = Worker::getCurrentWorkerId();
	std::atomic<SleeperState>& workerState = states[workerId];

	const SleeperState oldState = workerState.exchange(SleeperState::Sleeping);
	// Someone has notified us specifically -> skip sleeping.
	if (oldState == SleeperState::Notified) {
		workerState.store(SleeperState::Running, std::memory_order_relaxed);
		return;
	}

	// Me must have been running.
	assert(oldState == SleeperState::Running);

	// Decrement the global semaphore count.
	// See fetch_add comment in SpuriousFutex2Semaphore.hpp.
	CounterType c = counter.fetch_sub(1, std::memory_order_relaxed) - 1;
	assert(c >= -workerCount);
	LOGD("Decrement counter to: " << c);

	while (c < 0 && workerState.load(std::memory_order_relaxed) == SleeperState::Sleeping) {
		// NOLINTNEXTLINE(modernize-avoid-c-arrays)
		struct futex_waitv waiters[2];
		// Global futex
		init_futex_waitv(&waiters[0], c, &counter);

		// Specific futex
		init_futex_waitv(&waiters[1], static_cast<uint64_t>(SleeperState::Sleeping), &workerState);

		LOGD("Sleep on both futex");
		long ret = syscall(SYS_futex_waitv, waiters, 2, 0, nullptr, 0);
		if (ret == -1) {
			assert(errno == EAGAIN);
			c = getValue();
			continue;
		}
		// Someone called FUTEX_WAIT on either of the two futex

		// The futex(2) manpage says that applications should check the futex word
		// again because it is possible to have spurious wakeups.
		// We can't check the futex value because there is no way to decided
		// if someone called notify.
		// A sound generic Semaphore needs to split its atomic counter into
		// a semaphore value and a waiter count part.

		LOGD("futex_waitv returned with: " << ret);
		break;
	}
	workerState.store(SleeperState::Running, std::memory_order_relaxed);
}

}	 // namespace emper::lib::sync
