// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021 Florian Fischer
#pragma once

#include <linux/futex.h>
#include <sys/syscall.h>
#include <unistd.h>

#include <atomic>
#include <cstdint>
#include <new>
#include <ostream>

#include "Common.hpp"
#include "Debug.hpp"
#include "emper-common.h"
#include "lib/LinuxVersion.hpp"
#include "sleep_strategy/SleeperState.hpp"

namespace emper::lib::sync {

/**
 * @brief A semaphore implementation using two futexes to wait
 *
 * SpuriousFutex2Semaphore behaves like a normal Semaphore when using
 * only notify() and notify_many(count).
 * But it also supports notify_specific(workerId).
 * The notify_specific(workerId) is implemented with a exclusive futex per
 * worker. The worker specific futexes are a cache line exclusive SleeperState
 * per worker. A worker calling wait will use futex_waitv(2) introduced in linux
 * v5.16 to wait on both futexes.
 *
 * To notify a specific Worker we first set its SleeperState to Notified
 * and check the previous value:
 * - If it was running it will skip the next wait() and reset it
 * - If it was sleeping we have to call FUTEX_WAKE on the futex
 *   - The futex_waitv call will fail if the notifier races the futex_waitv call
 *   - If futex_waitv wins the race the notifier's FUTEX_WAKE will wake it
 *
 * ATTENTION: This semaphore implementation is prone to spurious wakeups!
 * SpuriousFutex2Semaphore::wait() may return spurious without a call to
 * SpuriousFutex2Semaphore::notify().
 * Use this class only where it is safe to have spurious wakeups.
 * It is designed to wake up and suspend worker threads it does not provided any
 * safety guaranties when protecting from data races.
 */
class SpuriousFutex2Semaphore : Logger<LogSubsystem::SLEEP_S> {
 public:
	// All futexes are currently 32 bits in size
	using CounterType = int32_t;

 private:
	using SleeperState = emper::sleep_strategy::SleeperState;
	// States of the worker threads
	std::atomic<SleeperState>* states;

	// >= 0 means no waiters
	// < 0 means there are waiters
	std::atomic<CounterType> counter;

	int workerCount;

 public:
	SpuriousFutex2Semaphore(CounterType counter = 0) : counter(counter) {
		if (EMPER_LINUX_LT("5.16.0"))
			DIE_MSG("SpuriousFutex2Semaphore needs futex_waitv(2) introduced in linux v5.16");
	}

	~SpuriousFutex2Semaphore() { delete states; }

	void init(workerid_t workerCount) {
		this->workerCount = workerCount;
		// Allcoate each worker specific SleeperState in its own cache line
		states = new (std::align_val_t(CACHE_LINE_SIZE)) std::atomic<SleeperState>[workerCount];
	}

	[[nodiscard]] inline auto getValue() const -> CounterType {
		return counter.load(std::memory_order_relaxed);
	}

	inline void notify_many(CounterType count) {
		// Using std::memory_order_relaxed is fine since we don't
		// give any memory synchronization guaranties because of the
		// possible spurious wakeups
		CounterType c = counter.fetch_add(count, std::memory_order_relaxed);
		LOGD("Futex2Sem: notify_many(" << count << ") inc counter to: " << c + count);
		if (c < 0) {
			LOGD("Futex2Sem: notify_many(" << count << ") wake " << count << " on " << &counter);
			syscall(SYS_futex, &counter, FUTEX_WAKE | FUTEX_PRIVATE_FLAG, count, 0, 0, 0);
		}
	}

	inline void notify() { notify_many(1); }

	inline void notifySpecific(workerid_t workerId) {
		std::atomic<SleeperState>& workerState = states[workerId];
		SleeperState oldState = workerState.exchange(SleeperState::Notified, std::memory_order_relaxed);

		// We are responsible to wake the futex
		if (oldState == SleeperState::Sleeping) {
			// The sleeper has decremented the semaphore count before sleeping.
			// We have to increment it when not using the counter to wakeup
			// to prevent the counter from shrinking uncontrolled.
			counter.fetch_add(1, std::memory_order_relaxed);
			LOGD("increment counter to " << counter << " and wake specific");
			syscall(SYS_futex, &workerState, FUTEX_WAKE | FUTEX_PRIVATE_FLAG, 1, 0, 0, 0);
		}
	}

	void wait();
};

}	 // namespace emper::lib::sync
