// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus, Florian Fischer
#include "log.hpp"

#include <pthread.h>

#include <chrono>
#include <ctime>		 // for localtime
#include <iomanip>	 // for operator<<, setfill, setw
#include <iostream>	 // for operator<<, cerr, ostream, basic_ostream
#include <mutex>		 // for mutex, unique_lock
#include <sstream>

#include "Common.hpp"
#include "Emper.hpp"
#include "Runtime.hpp"
#include "Worker.hpp"
#include "emper-common.h"	 // for workerid_t
#include "io/GlobalIoContext.hpp"
#include "log/LogBuffer.hpp"

using emper::io::GlobalIoContext;

namespace emper::log {

static void add_timestamp_to(std::ostringstream& logMessage) {
	using clock = std::chrono::system_clock;

	auto now = clock::now();
	auto now_time_t = clock::to_time_t(now);

	const std::tm* now_localtime = [&now_time_t] {
		if constexpr (emper::LOG_TIMESTAMP == emper::LogTimeStamp::utc) {
			std::tm utc_time;
			auto* res = gmtime_r(&now_time_t, &utc_time);
			if (unlikely(res != &utc_time)) DIE_MSG_ERRNO("gmtime_r failed");
			return res;

			// Using std::localtime takes a global lock to protect the tzinfo database
			// therefore using local timestamps is not scalable and therefore not useful
			// for finding timing sensible bugs.
		} else {
			return std::localtime(&now_time_t);
		}
	}();

	logMessage << std::put_time(now_localtime, "%H%M.%S");

	auto now_nanos = std::chrono::time_point_cast<std::chrono::nanoseconds>(now);
	auto time_since_epoch = now_nanos.time_since_epoch();
	long time_since_epoch_long = time_since_epoch.count();

	constexpr long NanosInASecond = 1000L * 1000 * 1000;
	long remaining_nanos = time_since_epoch_long % NanosInASecond;
	logMessage << " " << std::setfill('0') << std::setw(9) << remaining_nanos;
}

static std::mutex log_mutex;

void log(const std::string& prefix, const std::string& message) {
	Runtime* runtime = Runtime::getRuntime();
	Worker* worker = Worker::getCurrentWorker();
	std::ostringstream logMessage;
	if (likely(worker)) {
		workerid_t workerId = worker->getWorkerId();
		std::string workerIdAsString = std::to_string(workerId);
		logMessage << std::setfill('0') << std::setw(3) << workerIdAsString << " ";
	} else {
		// Are we the global IO completer
		GlobalIoContext* gio = runtime ? runtime->globalIo : nullptr;
		if (gio && pthread_self() == gio->completer) {
			logMessage << "IOC ";
		} else {
			logMessage << "    ";
		}
	}

	if constexpr (emper::LOG_TIMESTAMP != emper::LogTimeStamp::none) {
		add_timestamp_to(logMessage);
	}

	if (!prefix.empty()) {
		logMessage << " " << prefix << " ";
	} else {
		logMessage << "       ";
	}

	logMessage << message << std::endl;

	if (likely(runtime) && runtime->logBuffer) {
		runtime->logBuffer->log(logMessage.str());
	} else {
		std::unique_lock<std::mutex> lock(log_mutex);
		std::cerr << logMessage.str();
	}
}
}	 // namespace emper::log
