// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2022 Florian Fischer
#include "sleep_strategy/AbstractIoSleepStrategy.hpp"

#include <liburing.h>
#include <liburing/io_uring.h>

#include <atomic>
#include <cassert>
#include <cstdint>
#include <new>
#include <vector>

#include "CallerEnvironment.hpp"
#include "Common.hpp"
#include "Emper.hpp"
#include "Runtime.hpp"
#include "Worker.hpp"
#include "io/IoContext.hpp"
#include "lib/TaggedPtr.hpp"

using emper::io::IoContext;
using emper::lib::TaggedPtr;

namespace emper::sleep_strategy {

thread_local bool AbstractIoSleepStrategy::readingGlobal = false;

AbstractIoSleepStrategy::AbstractIoSleepStrategy(Runtime& runtime, workerid_t workerCount)
		: workerCount(workerCount), stats(runtime), runtime(runtime) {
	sleepers.store(0, std::memory_order_relaxed);

	sleepStates = new (std::align_val_t(CACHE_LINE_SIZE)) std::atomic<SleeperState>[workerCount];
}

AbstractIoSleepStrategy::~AbstractIoSleepStrategy() { delete[] sleepStates; }

auto AbstractIoSleepStrategy::prepareSleep(const std::function<bool(void)>& customLogic) -> bool {
	const workerid_t workerId = Worker::getCurrentWorkerId();
	std::atomic<SleeperState>& sleepState = sleepStates[workerId];

	// Check if we were notified specifically and should skip the sleep attempt
	const SleeperState oldState = sleepState.exchange(SleeperState::Sleeping);
	// Someone has notified us specifically -> skip sleeping.
	if (oldState == SleeperState::Notified) {
		LOGD("State was notified -> reset notified state to running and skip sleeping");
		goto skip;
	}

	// Me must have been running or already reading from the global IO object.
	assert(oldState == SleeperState::Running || readingGlobal);

	if (!readingGlobal) {
		const bool shouldSleep = customLogic();
		if (!shouldSleep) goto skip;
		readingGlobal = true;
	}

	return true;

skip:
	sleepState.store(SleeperState::Running, std::memory_order_relaxed);
	return false;
}

void AbstractIoSleepStrategy::submitAndWaitForCompletionsIn(IoContext& io) {
	LOGD("sleep until IO completions occur");
	io.submitPreparedSqesAndWait<CallerEnvironment::EMPER>(1);
	LOGD("wakeup due to available IO completions");

	stats.incWakeup();
}

void AbstractIoSleepStrategy::prepareSleepRead(IoContext& io, int fd, uint8_t sqe_flags) {
	struct io_uring_sqe* sqe = io.getSqe();

	// We read directly into the workers dispatchHint
	// TODO: Think about the race between the two reads
	io_uring_prep_read(sqe, fd, &io.worker->dispatchHint, sizeof(io.worker->dispatchHint), 0);

	sqe->flags = sqe_flags;

	io_uring_sqe_set_data(
			sqe, TaggedPtr((void*)nullptr,
										 static_cast<uint16_t>(IoContext::PointerTags::NewWorkNotification)));

	io.trackReqsInUring(1);
}

void AbstractIoSleepStrategy::sendSpecificNotification(workerid_t workerId) {
	stats.addNotifications(1);

	emper::assertInRuntime();
	auto& io = *IoContext::getWorkerIo();
	struct io_uring_sqe* sqe = io.getSqe();

	TaggedPtr tag((void*)nullptr, static_cast<uint16_t>(IoContext::PointerTags::NewWorkNotification),
								true);

	io_uring_prep_msg_ring(sqe, runtime.ioContexts[workerId]->ring.ring_fd, 0, uintptr_t(tag), 0);

	// Do not generate a cqe on the submiting io_uring
	sqe->flags = IOSQE_CQE_SKIP_SUCCESS;
	io_uring_sqe_set_data(sqe, tag);

	// The sent message will create a cqes one on the messaged io_uring
	runtime.ioContexts[workerId]->trackReqsInUring(1);
	io.submitPreparedSqesAndWait<CallerEnvironment::EMPER>(0);
}

template <CallerEnvironment callerEnvironment>
void AbstractIoSleepStrategy::onNewWorkNotification(IoContext& io,
																										const emper::lib::TaggedPtr& data) {
	if constexpr (callerEnvironment != CallerEnvironment::OWNER) {
		DIE_MSG("Others reaping new work notification from " << io.worker->getWorkerId() << " CQ");
	}

	assert(data.getTag() == static_cast<uint16_t>(IoContext::PointerTags::NewWorkNotification));

	stats.incWakeupDueToNotify();

	std::atomic<SleeperState>& sleepState = sleepStates[io.worker->getWorkerId()];

	if (data.isMarked()) {
		LOGD("Got specific notification");
	} else {
		LOGD("Got new work notification");
		// Reset global flag to indicate that a new sleep cqe must be prepared
		readingGlobal = false;
	}

	sleepState.store(SleeperState::Running, std::memory_order_release);
}

template void AbstractIoSleepStrategy::onNewWorkNotification<CallerEnvironment::OWNER>(
		IoContext& io, const TaggedPtr& data);
template void AbstractIoSleepStrategy::onNewWorkNotification<CallerEnvironment::EMPER>(
		IoContext& io, const TaggedPtr& data);
template void AbstractIoSleepStrategy::onNewWorkNotification<CallerEnvironment::ANYWHERE>(
		IoContext& io, const TaggedPtr& data);

}	 // namespace emper::sleep_strategy
