// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2022 Florian Fischer
#pragma once

#include <algorithm>
#include <atomic>
#include <cassert>
#include <cstdint>
#include <functional>
#include <iostream>

#include "CallerEnvironment.hpp"
#include "Common.hpp"
#include "Debug.hpp"
#include "FiberHint.hpp"
#include "emper-common.h"
#include "lib/TaggedPtr.hpp"
#include "sleep_strategy/SleeperState.hpp"
#include "sleep_strategy/Stats.hpp"

class Runtime;

namespace emper::io {
class IoContext;
}

namespace emper::sleep_strategy {
/**
 * @brief An abstract IO based sleep strategy combining the IO subsystem with the sleep strategy
 *
 * Design goals
 * ============
 *
 * * Wakeup either on external new work notifications or on local IO completions
 *   -> Sleep strategy is sound without the IO completer
 * * Do as less as possible in a system saturated with work
 * * Pass a hint where to find new work to suspended workers
 *
 * Algorithm
 * =========
 *
 *```
 * Data:
 *     Global:
 *             global IO object
 *             sleepers count
 *     Per worker:
 *             dispatch hint buffer
 *             readingGlobal flag
 *             atomic sleep state
 *
 * PrepareSleep(customLogic):
 *     Exchange state with sleeping
 *     If state was notified
 *         Reset to running
 *         Return
 *     If not readingGlobal
 *         Return false if customLogic returns false
 *         Set readingGlobal
 *     Return true
 *
 * WriteNotifications : implementation specific
 *
 * NotifyEmper(n):
 *    if observed sleepers <= 0
 *            return
 *
 *    // Determine how many we are responsible to wake
 *    do
 *            toWakeup = min(observed sleepers, n)
 *    while (!CAS(sleepers, toWakeup))
 *
 *    write toWakeup hints
 *
 * NotifyAnywhere(n):
 *    // Ensure all n notifications take effect
 *    while (!CAS(sleepers, observed sleepers - n))
 *        if observed sleeping <= -n
 *            return
 *
 *    toWakeup = min(observed sleeping, n)
 *    write toWakeup hints
 *
 * NotifySpecific(w):
 *    Get w's state
 *    Relaxed exchange with notified
 *    Return w was sleeping
 *        Send specific notification
 *
 * onNewWorkCompletion(data):
 *     Release state to running
 *     If data is marked
 *         assert was notified
 *     Else
 *         Reset readingGlobal
 *```
 *
 * Notes
 * =====
 *
 * * We must decrement the sleepers count on the notifier side to
 *   prevent multiple notifiers to observe all the same amount of sleepers,
 *   trying to wake up the same sleepers by writing to the e.g a pipe will jam
 *   it up with unconsumed hints and thus block in writeNotifications resulting
 *   in a deadlock.
 * * The CAS loops on the notifier side are needed because decrementing
 *   and incrementing is racy: Two notifier can observe the
 *   sum of both their excess decrement and increment to much resulting in a
 *   broken counter.
 * * Add the dispatch hint code in AbstractWorkStealingScheduler::nextFiber.
 *   This allows workers to check the dispatch hint after there
 *   where no local work to execute.
 *   This is a trade-off where we trade slower wakeup - a just awoken worker
 *   will check for local work - against a faster dispatch hot path when
 *   we have work to do in our local WSQ.
 * * Other thread must not reap new work notifications because this
 *   would introduces a race for cqes and a possible
 *   lost wakeup if the other consumes the completions before the worker
 *   is actually waiting for them.
 * * When notifying sleeping workers from anywhere we must ensure that all
 *   notifications take effect. This is needed for example when terminating
 *   the runtime to prevent sleep attempt from worker thread which are
 *   about to sleep but have not incremented the sleeper count yet.
 *   We achieve this by always decrementing the sleeper count by the notification
 *   count.
 */
class AbstractIoSleepStrategy : public Logger<LogSubsystem::SLEEP_S> {
	friend class emper::io::IoContext;

 protected:
	/**
	 * @brief decrement the sleep count and return how many workers should be notified
	 *
	 * As mentioned in the AbstractIoSleepStrategy documentation We need to decrement
	 * the sleeper count on the notification side to prevent overcommiting notifications.
	 * And the CAS loop is needed because we don't want to decrement the sleepers
	 * count to much.
	 *
	 * Decrement the sleeper count for each sleeper we are going to wakeup
	 */
	virtual auto getSleeperCountToWakeupEmper(int64_t& sleeping, int64_t count) -> uint64_t {
		int64_t toWakeup;
		do {
			// No one is sleeping -> we don't need to notify
			if (sleeping <= 0) {
				return 0;
			}

			// Count of sleepers we notify
			toWakeup = std::min(sleeping, count);
		} while (!sleepers.compare_exchange_weak(sleeping, sleeping - toWakeup,
																						 std::memory_order_release, std::memory_order_acquire));

		LOGD("notifyFromEmper writes "
				 << toWakeup << " notifications and set sleepers count to: " << sleeping - toWakeup);

		assert(toWakeup >= 0);
		return static_cast<uint64_t>(toWakeup);
	}

	/**
	 * @brief decrement the sleep count and return how many workers should be notified
	 *
	 * If we observe nobody sleeping we need to prevent sleep locks when
	 * everybody is about to sleep but none has incremented the sleepers count yet.
	 * We prevent the next sleep from blocking by setting the sleeper count to
	 * the negative amount we want to prevent from sleeping.
	 *
	 * More general if we notify from anywhere we need all notifications to take
	 * effect. Only notifying the sleepers we observe may not be enough.
	 * For example when terminating the runtime we notify all workers if we
	 * only wakeup the ones we observe sleeping we may never terminate because
	 * after we notified all we observed new one went to sleep which will never
	 * be notified.
	 * To prevent this we make sure that all count notifications take effect
	 * by making the sleepers count negative and thus preventing at least count
	 * sleep attempts.
	 */
	virtual auto getSleeperCountToWakeupAnywhere(int64_t& sleeping, int64_t count) -> uint64_t {
		int64_t newSleeping;
		do {
			// We already prevent enough sleep attempts
			if (sleeping <= -count) {
				LOGD("notifyFromAnywhere sleeper count already preventing enough sleep attempts");
				return 0;
			}

			// Don't decrease the sleeper count in the CAS loop further than -count,
			// which is the threshold we need to ensure that the notifications will be
			// observed.
			// Decreasing it further than this threshold is not faulty it just
			// results in unnecessary skipped sleeps.
			newSleeping = std::max(-count, sleeping - count);
		} while (!sleepers.compare_exchange_weak(sleeping, newSleeping, std::memory_order_release,
																						 std::memory_order_acquire));

		// toWakeup should not be negative but could be if sleeping was already negative
		const int64_t toWakeup = std::max(std::min(sleeping, count), 0L);
		LOGD("notifyFromAnywhere writes "
				 << toWakeup << " notifications and set sleepers count to: " << newSleeping);

		assert(toWakeup >= 0);
		return static_cast<uint64_t>(toWakeup);
	}

	static thread_local bool readingGlobal;
	std::atomic<SleeperState>* sleepStates;

	// Make sure the shared counter lives in an exclusive cache line
	CACHE_LINE_EXCLUSIVE(std::atomic<int64_t>, sleepers);

	workerid_t workerCount;
	Stats stats;

	const Runtime& runtime;

	AbstractIoSleepStrategy(Runtime& runtime, workerid_t workerCount);
	virtual ~AbstractIoSleepStrategy();

	virtual void writeNotificationsFromEmper(emper::FiberHint hint, unsigned count) = 0;
	virtual void writeNotificationsFromAnywhere(emper::FiberHint hint, unsigned count) = 0;

	void sendSpecificNotification(workerid_t workerId);

	using CustomLogic = std::function<bool(void)>;

	/**
	 * @brief Prepare everything to sleep until new work is available
	 *
	 * @param customLogic Code specific to the actual IO sleep strategy. Reads are only prepared
	 *                    if customLogic returns true.
	 *
	 * @return True if any read was prepared and IO completions should be awaited
	 */
	auto prepareSleep(const CustomLogic& customLogic) -> bool;

	/**
	 * @brief prepare a new work notification read request
	 *
	 * @param io The IoContext where the read request will be prepared
	 * @param fd The file descriptor to read from
	 * @param sqe_flags The flags set on the issued SQE
	 */
	static void prepareSleepRead(emper::io::IoContext& io, int fd, uint8_t sqe_flags = 0);

	void submitAndWaitForCompletionsIn(emper::io::IoContext& io);

	template <CallerEnvironment callerEnvironment>
	void onNewWorkNotification(emper::io::IoContext& io, const emper::lib::TaggedPtr& data);

	template <CallerEnvironment callerEnvironment>
	auto getSleeperCountToWakeup(int64_t& sleeping, unsigned count) -> unsigned {
		auto signedCount = static_cast<int64_t>(count);

		if constexpr (callerEnvironment == CallerEnvironment::EMPER) {
			return getSleeperCountToWakeupEmper(sleeping, signedCount);
		} else {
			return getSleeperCountToWakeupAnywhere(sleeping, signedCount);
		}
	}

	template <CallerEnvironment callerEnvironment>
	void notify(emper::FiberHint hint, unsigned count) {
		stats.incOnNewWork();

		int64_t sleeping = getSleeping();

		const unsigned toWakeup = getSleeperCountToWakeup<callerEnvironment>(sleeping, count);
		if (toWakeup) {
			stats.incNotify();
			stats.addNotifications(toWakeup);
			if constexpr (callerEnvironment == CallerEnvironment::EMPER) {
				writeNotificationsFromEmper(hint, toWakeup);
			} else {
				writeNotificationsFromAnywhere(hint, toWakeup);
			}
		}
	}

	template <CallerEnvironment callerEnvironment>
	void notifySpecificInternal(workerid_t workerId) {
		auto& specificState = sleepStates[workerId];

		LOGD("Specifically notify  worker " << workerId << " from " << callerEnvironment);

		auto oldState = specificState.exchange(SleeperState::Notified, std::memory_order_relaxed);
		// We are responsible to wake the futex
		if (oldState == SleeperState::Sleeping) {
			LOGD(workerId << " was sleeping send notification");
			sendSpecificNotification(workerId);
		}
	}

 public:
	void printStats(std::ostream& out = std::cout) { stats.print(out); }

	[[nodiscard]] inline auto getSleeping() const -> long { return sleepers.load(); }

	virtual void sleep() = 0;
};
}	 // namespace emper::sleep_strategy
