// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021-2022 Florian Fischer
#pragma once

#include <sys/types.h>
#include <unistd.h>

#include <iostream>
#include <vector>

#include "CallerEnvironment.hpp"
#include "Common.hpp"
#include "Debug.hpp"
#include "FiberHint.hpp"
#include "emper-common.h"
#include "sleep_strategy/AbstractIoSleepStrategy.hpp"
#include "sleep_strategy/AbstractWorkerSleepStrategy.hpp"

class Runtime;

namespace emper::sleep_strategy {
/**
 * @brief A pipe(2) based sleep strategy combining the IO subsystem with the sleep strategy
 *
 * implement a pipe based sleep strategy using the IO subsystem
 */
class PipeSleepStrategy : public AbstractWorkerSleepStrategy<PipeSleepStrategy>,
													public AbstractIoSleepStrategy {
	class Pipe {
		friend class PipeSleepStrategy;
		int sleepFd;
		int notifyFd;

		Pipe() {
			// NOLINTNEXTLINE(modernize-avoid-c-arrays)
			int fds[2];
			if (::pipe(fds)) DIE_MSG_ERRNO("pipe failed");
			sleepFd = fds[0];
			notifyFd = fds[1];
		}

		~Pipe() {
			if (close(sleepFd) || close(notifyFd)) DIE_MSG("Closing pipe failed");
		}
	};

	Pipe global;

	void writeNotifications(emper::FiberHint hint, unsigned count) {
		std::vector<void*> hints(count, hint);

		ssize_t res = write(global.notifyFd, hints.data(), sizeof(void*) * hints.size());
		if (unlikely(res) < 0) {
			DIE_MSG_ERRNO("writing to notifyFd failed");
		}
	}

	void writeNotificationsFromEmper(emper::FiberHint hint, unsigned count) override {
		writeNotifications(hint, count);
	}
	void writeNotificationsFromAnywhere(emper::FiberHint hint, unsigned count) override {
		writeNotifications(hint, count);
	}

 public:
	PipeSleepStrategy(Runtime& runtime, workerid_t workerCount)
			: AbstractIoSleepStrategy(runtime, workerCount) {}

	void printStats(std::ostream& out = std::cout) { AbstractIoSleepStrategy::printStats(out); }

	template <CallerEnvironment callerEnvironment>
	inline void notifyOne() {
		notifyMany<callerEnvironment>(1);
	}

	template <CallerEnvironment callerEnvironment>
	inline void notifyMany(unsigned count) {
		if (count > 1) LOGD("notify(hint, " << count << ")");
		notify<callerEnvironment>(emper::FiberHint::createNewWorkHint<callerEnvironment>(), count);
	}

	template <CallerEnvironment callerEnvironment>
	inline void notifyAll() {
		notifyMany<callerEnvironment>(workerCount);
	}

	template <CallerEnvironment callerEnvironment>
	inline void notifySpecific(workerid_t workerId) {
		notifySpecificInternal<callerEnvironment>(workerId);
	}

	void sleep() override;
};
}	 // namespace emper::sleep_strategy
