// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021 Florian Fischer
#include "sleep_strategy/WaitfdSleepStrategy.hpp"

#include <liburing.h>
#include <liburing/io_uring.h>
#include <sys/syscall.h>

#include <functional>

#include "Emper.hpp"
#include "FiberSource.hpp"
#include "Runtime.hpp"
#include "io/IoContext.hpp"

using emper::io::IoContext;

#ifndef IOURING_REGISTER_WAITFD
static const unsigned IOURING_REGISTER_WAITFD = []() -> unsigned {
	if (EMPER_LINUX_GE("5.19")) return 24;
	if (EMPER_LINUX_GE("5.18")) return 22;
	return 20;
}();
#endif

struct io_uring_waitfd_registration {
	void* hint;
	int fd;
};

namespace emper::sleep_strategy {

uint32_t WaitfdSleepStrategy::WAITFD_NOTIFY_AND_PREVENT_SLEEP = 1;

WaitfdSleepStrategy::WaitfdSleepStrategy(Runtime& runtime, workerid_t workerCount,
																				 unsigned waitfdFlags)
		: AbstractIoSleepStrategy(runtime, workerCount), waitfdFlags(waitfdFlags) {
	long res = syscall(__NR_waitfd, &sleepers, waitfdFlags);
	if (res < 0) {
		DIE_MSG_ERRNO("creating waitfd failed");
	}
	waitfd = static_cast<int>(res);

	if constexpr (emper::IO_NOTIFICATION) {
		auto registerWaitFd = [this](workerid_t workerId) {
			IoContext& io = *IoContext::getWorkerIo();
			emper::FiberHint hint(workerId, emper::FiberSource::hintIo);
			struct io_uring_waitfd_registration reg {
				hint, this->waitfd
			};
			long ret = syscall(__NR_io_uring_register, io.ring.ring_fd, IOURING_REGISTER_WAITFD, &reg,
												 sizeof(reg));
			if (ret) {
				DIE_MSG_ERRNO("failed to register waitfd for worker " << workerId);
			}
		};
		runtime.addNewWorkerHook(registerWaitFd);
	}
}

WaitfdSleepStrategy::~WaitfdSleepStrategy() { close(waitfd); }

void WaitfdSleepStrategy::sleep() {
	IoContext& io = *IoContext::getWorkerIo();

	auto customLogic = [this, &io] {
		// waitfd non-blocking reads will always return -EAGAIN and waitfd is not pollable
		// therefore we can skip that and immediately issue the read in an async manner.
		uint8_t sqe_flags = 0;
		if constexpr (emper::IO_PUNT_WAITFD_READ) sqe_flags |= IOSQE_ASYNC;
		prepareSleepRead(io, waitfd, sqe_flags);
		LOGD("prepared waitfd read");
		return true;
	};
	const bool shouldSleep = prepareSleep(customLogic);

	if (!shouldSleep) return;

	submitAndWaitForCompletionsIn(io);
}
}	 // namespace emper::sleep_strategy
