// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021 Florian Fischer
#pragma once

#include <sys/types.h>
#include <unistd.h>

#include <algorithm>
#include <cstdint>
#include <cstdlib>
#include <iostream>
#include <string>

#include "CallerEnvironment.hpp"
#include "Common.hpp"
#include "Debug.hpp"
#include "FiberHint.hpp"
#include "emper-common.h"
#include "emper-config.h"	 // IWYU pragma: keep
#include "lib/LinuxVersion.hpp"
#include "sleep_strategy/AbstractIoSleepStrategy.hpp"
#include "sleep_strategy/AbstractWorkerSleepStrategy.hpp"

class Runtime;

#ifndef __NR_waitfd
static const unsigned __NR_waitfd = []() -> unsigned {
	if (EMPER_LINUX_GE("5.17"))
		return 451;
	else if (EMPER_LINUX_GE("5.16"))
		return 450;
	else
		return 449;
}();
#endif

namespace emper::sleep_strategy {

class WaitfdSleepStrategy : public AbstractWorkerSleepStrategy<WaitfdSleepStrategy>,
														public AbstractIoSleepStrategy {
	int waitfd;

	enum class WaitfdFlags {
		none = 0,
		accounting = 1,
		waitFree = 2,
		bestEffort = 4,
	};

	const unsigned waitfdFlags;

	struct WaitfdNotification {
		void* d;
		uint32_t flags;
		uint32_t count;
	};

	static auto getDefaultWaitfdFlags() -> unsigned {
		unsigned flags = 0;
		flags = static_cast<unsigned>(WaitfdFlags::EMPER_WAITFD_FLAGS);

		const auto* flagsEnvVal = std::getenv("EMPER_WAITFD_FLAGS");
		if (flagsEnvVal) {
			const std::string envValStr(flagsEnvVal);
			if (envValStr == "accounting")
				flags = static_cast<unsigned>(WaitfdFlags::accounting);
			else if (envValStr == "wait-free")
				flags = static_cast<unsigned>(WaitfdFlags::waitFree);
			else if (envValStr == "best-effort")
				flags = static_cast<unsigned>(WaitfdFlags::bestEffort);
			else
				DIE_MSG("Unknown waitfd flags specified in environment: " << envValStr);
		}

		return flags;
	}

	void writeNotifications(emper::FiberHint hint, uint32_t flags, int32_t count) {
		struct WaitfdNotification notification = {hint, flags, static_cast<uint32_t>(count)};

		LOGD("write: {" << notification.d << ", " << notification.flags << ", " << notification.count
										<< "}");

		ssize_t res = write(waitfd, reinterpret_cast<void*>(&notification), sizeof(notification));
		if (unlikely(res) < 0) {
			DIE_MSG_ERRNO("writing to waitfd failed");
		}
	}

	static uint32_t WAITFD_NOTIFY_AND_PREVENT_SLEEP;

	void writeNotificationsFromEmper(emper::FiberHint hint, unsigned count) override {
		writeNotifications(hint, 0, count);
	}

	void writeNotificationsFromAnywhere(emper::FiberHint hint, unsigned count) override {
		writeNotifications(hint, WAITFD_NOTIFY_AND_PREVENT_SLEEP, count);
	}

	// If the waitfd implementation does the sleeper accounting in kernel space
	// we must not decrement the sleeper count in userspace.
	auto getSleeperCountToWakeupEmper(int64_t& sleeping, int64_t count) -> uint64_t override {
		if (waitfdFlags != static_cast<unsigned>(WaitfdFlags::accounting))
			return AbstractIoSleepStrategy::getSleeperCountToWakeupEmper(sleeping, count);

		const int64_t toWakeup = std::min(sleeping, count);
		LOGD("notifyFromEmper writes " << toWakeup << " notifications");
		return toWakeup > 0 ? static_cast<uint64_t>(toWakeup) : 0;
	}

	auto getSleeperCountToWakeupAnywhere(int64_t& sleeping, int64_t count) -> uint64_t override {
		if (waitfdFlags != static_cast<unsigned>(WaitfdFlags::accounting))
			return AbstractIoSleepStrategy::getSleeperCountToWakeupAnywhere(sleeping, count);

		if (sleeping <= -count) {
			LOGD("notifyFromAnywhere sleeper count already preventing enough sleep attempts");
			return 0;
		}
		LOGD("notifyFromAnywhere writes " << count << " notifications");
		return static_cast<uint64_t>(count);
	}

 public:
	WaitfdSleepStrategy(Runtime& runtime, workerid_t workerCount,
											unsigned waitfdFlags = getDefaultWaitfdFlags());
	~WaitfdSleepStrategy();

	template <CallerEnvironment callerEnvironment>
	inline void notifyOne() {
		notifyMany<callerEnvironment>(1);
	}

	template <CallerEnvironment callerEnvironment>
	inline void notifyMany(unsigned count) {
		if (count > 1) LOGD("notify(hint, " << count << ")");
		notify<callerEnvironment>(emper::FiberHint::createNewWorkHint<callerEnvironment>(), count);
	}

	template <CallerEnvironment callerEnvironment>
	inline void notifyAll() {
		notifyMany<callerEnvironment>(workerCount);
	}

	template <CallerEnvironment callerEnvironment>
	inline void notifySpecific(ATTR_UNUSED workerid_t workerId) {
		notifySpecificInternal<callerEnvironment>(workerId);
	}

	void printStats(std::ostream& out = std::cout) { AbstractIoSleepStrategy::printStats(out); }
	void sleep() override;
};
}	 // namespace emper::sleep_strategy
