// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2022 Florian Schmaus
#include <algorithm>
#include <boost/optional.hpp>
#include <boost/program_options.hpp>
#include <cassert>
#include <chrono>
#include <compare>
#include <cstdint>
#include <cstdlib>
#include <exception>
#include <filesystem>
#include <fstream>
#include <functional>
#include <iostream>
#include <numeric>
#include <optional>
#include <random>
#include <ratio>	// IWYU pragma: keep
#include <string>
#include <vector>

#include "CountingPrivateSemaphore.hpp"
#include "Fibril.hpp"
#include "Runtime.hpp"
#include "emper-common.h"
#include "emper.hpp"
#include "lib/math.hpp"

namespace chrn = std::chrono;
namespace fs = std::filesystem;
namespace math = emper::lib::math;
namespace po = boost::program_options;

using Clock = std::chrono::high_resolution_clock;

static double utilization = 0.8;
static std::uint64_t pulse_target_work_ns = 233'000;
static unsigned int pulse_concurrency_level = 4;
static std::uint_fast64_t seed = 20141007;
static unsigned int duration_ms = 10'000;
static workerid_t worker_count = Runtime::getDefaultWorkerCount();
static bool trace = false;
static boost::optional<fs::path> result_dir;

// Derived values from the user supplied configuration values.
static std::uint64_t duration_ns;
static std::uint64_t work_duration_ns;
static std::uint64_t sleep_duration_ns;
static std::uint64_t number_of_pulses;
static std::uint64_t pulse_target_sleep_ns;

struct PulseTrace {
	std::uint64_t work_ns;
	chrn::time_point<Clock> work_start;
	chrn::time_point<Clock> work_finish;
	std::uint64_t sleep_ns;
	chrn::time_point<Clock> sleep_start;
};

class Pulser {
 private:
	std::mt19937_64 rng;

	std::uint64_t remaining_work_ns;
	std::uint64_t remaining_sleep_ns;

 public:
	std::vector<PulseTrace> traces;
	chrn::time_point<Clock> pulse_start;

 private:
	static void busyWait(std::uint64_t work_ns) {
		const auto deadline = Clock::now() + std::chrono::nanoseconds(work_ns);
		while (Clock::now() < deadline) {
		}
	}

	// NOLINTNEXTLINE(clang-diagnostic-unknown-attributes)
	emper_fibril static void work(std::uint64_t work_ns) {
		if (pulse_concurrency_level) {
			std::uint64_t work_per_concurrency_level_ns = work_ns / pulse_concurrency_level;

			Fibril fibril;
			for (unsigned int i = 0; i < pulse_concurrency_level; ++i) {
				fibril.spawn(&Pulser::busyWait, work_per_concurrency_level_ns);
			}
			fibril.sync();
		} else {
			busyWait(work_ns);
		}
	}

	auto getNanos(std::binomial_distribution<std::uint64_t>& dist, std::uint64_t remaining_ns,
								std::uint64_t target_ns, std::uint64_t pulse_number) -> std::uint64_t {
		assert(remaining_ns);

		std::uint64_t nanos = dist(rng);

		if (nanos > remaining_ns) {
			// If the random distribution returned a value higher than the
			// remaining value, then we need to cap the value.
			nanos = remaining_ns;

			if (nanos > target_ns) {
				std::uint64_t remaining_pulses = number_of_pulses - pulse_number - 1;
				nanos = nanos / remaining_pulses;
			}
		}

		return nanos;
	}

 public:
	Pulser(std::uint_fast64_t seed)
			: rng(seed), remaining_work_ns(work_duration_ns), remaining_sleep_ns(sleep_duration_ns) {
		if (trace) {
			traces = std::vector<PulseTrace>(number_of_pulses);
		}
	};

	void pulse() {
		std::binomial_distribution<std::uint64_t> work_dist(2 * pulse_target_work_ns, 0.5);
		std::binomial_distribution<std::uint64_t> sleep_dist(2 * pulse_target_sleep_ns, 0.5);

		pulse_start = Clock::now();
		for (std::uint64_t i = 0; i < number_of_pulses; ++i) {
			const auto work_ns = getNanos(work_dist, remaining_work_ns, pulse_target_work_ns, i);
			if (trace) {
				traces[i].work_start = Clock::now();
				traces[i].work_ns = work_ns;
			}
			work(work_ns);
			if (trace) traces[i].work_finish = Clock::now();
			remaining_work_ns -= work_ns;

			const auto sleep_ns = getNanos(sleep_dist, remaining_sleep_ns, pulse_target_sleep_ns, i);
			if (trace) {
				traces[i].sleep_ns = sleep_ns;
				traces[i].sleep_start = Clock::now();
			}
			emper::nanosleep(sleep_ns);
			// TODO: record actual sleep finish?
			remaining_sleep_ns -= sleep_ns;
		}
	}
};

static void bfp9000() {
	duration_ns = duration_ms * 1'000'000L;
	work_duration_ns = static_cast<std::uint64_t>(static_cast<double>(duration_ns) * utilization);
	sleep_duration_ns = duration_ns - work_duration_ns;
	number_of_pulses = work_duration_ns / pulse_target_work_ns;
	pulse_target_sleep_ns = sleep_duration_ns / number_of_pulses;

	std::mt19937_64 rng(seed);
	std::vector<Pulser> pulsers;
	pulsers.reserve(worker_count);
	for (auto i = 0; i < worker_count; ++i) {
		pulsers.emplace_back(rng());
	}

	// clang-format off
	std::cout << std::boolalpha
						<< "        *** BFP 9K ***" << std::endl
						<< "   The Big Fancy Pulser 9000" << std::endl
						<< std::endl
						<< "duration-ns: " << duration_ns << std::endl
						<< "work-duation-ns: " << work_duration_ns << std::endl
						<< "sleep-duration-ns: " << sleep_duration_ns << std::endl
						<< "pulse-target-work-ns: " << pulse_target_work_ns << std::endl
						<< "pulse-target-sleep-ns: " << pulse_target_sleep_ns << std::endl
						<< "number-of-pulses: " << number_of_pulses << std::endl
						<< "pulse-concurrency-level: " << pulse_concurrency_level << std::endl
						<< "worker-count: " << worker_count << std::endl
						<< "trace: " << trace << std::endl
						<< "seed: " << seed << std::endl
	;
	// clang-format on

	const auto start = Clock::now();
	{
		Runtime runtime(worker_count);
		runtime.executeAndWait([&pulsers] {
			CPS cps;
			for (auto& pulser : pulsers) {
				spawn([&pulser] { pulser.pulse(); }, cps);
			}
			cps.wait();
		});
	}
	const auto end = Clock::now();
	const auto actual_duration_us = chrn::duration_cast<chrn::microseconds>(end - start);

	// clang-format off
	std::cout << std::endl
						<< "actual-duration-us: " << actual_duration_us.count() << " us" << std::endl;
	// clang-format on

	if (!trace) return;

	std::uint64_t actual_total_work_ns = 0;
	std::uint64_t actual_total_sleep_ns = 0;
	std::vector<chrn::nanoseconds> latencies;
	latencies.reserve(pulsers.size() * number_of_pulses);

	std::optional<std::ofstream> latencies_file_strm;
	std::optional<std::ofstream> latencies_sorted_file_strm;
	std::optional<std::ofstream> work_file_strm;
	std::optional<std::ofstream> sleep_file_strm;
	std::optional<std::ofstream> csv_file_strm;
	if (result_dir) {
		auto latencies_file = *result_dir / "latencies";
		latencies_file_strm = std::ofstream(latencies_file);

		auto latencies_sorted_file = *result_dir / "latencies-sorted";
		latencies_sorted_file_strm = std::ofstream(latencies_sorted_file);

		auto work_file = *result_dir / "work";
		work_file_strm = std::ofstream(work_file);

		auto sleep_file = *result_dir / "sleep";
		sleep_file_strm = std::ofstream(sleep_file);

		auto csv_file = *result_dir / "csv";
		csv_file_strm = std::ofstream(csv_file);
	}

	for (auto& pulser : pulsers) {
		std::uint64_t pulser_work_ns = 0;
		std::uint64_t pulser_sleep_ns = 0;

		auto previous_sleep_finish = pulser.pulse_start;
		for (auto& trace : pulser.traces) {
			pulser_work_ns += trace.work_ns;
			pulser_sleep_ns += trace.sleep_ns;

			auto pulse_latency =
					chrn::duration_cast<chrn::nanoseconds>(trace.work_start - previous_sleep_finish);

			assert(pulse_latency.count());
			latencies.push_back(pulse_latency);

			if (!result_dir) continue;

			auto work_wall_clock =
					chrn::duration_cast<chrn::nanoseconds>(trace.work_finish - trace.work_start);
			auto work_speedup =
					static_cast<double>(trace.work_ns) / static_cast<double>(work_wall_clock.count());

			// NOLINTBEGIN(bugprone-unchecked-optional-access)
			*latencies_file_strm << pulse_latency.count() << std::endl;
			*work_file_strm << trace.work_ns << std::endl;
			*sleep_file_strm << trace.sleep_ns << std::endl;
			// clang-format off
			*csv_file_strm << previous_sleep_finish
										 << ", " << trace.work_start
										 << ", " << pulse_latency
										 << ", " << trace.work_finish
										 << ", " << work_wall_clock
										 << ", " << work_speedup
										;
			// clang-format on
			// NOLINTEND(bugprone-unchecked-optional-access)

			previous_sleep_finish = trace.work_finish + chrn::nanoseconds(trace.sleep_ns);
		}

		actual_total_work_ns += pulser_work_ns;
		actual_total_sleep_ns += pulser_sleep_ns;
	}

	std::sort(latencies.begin(), latencies.end());
	// NOLINTBEGIN(bugprone-unchecked-optional-access)
	if (latencies_sorted_file_strm)
		for (auto l : latencies) *latencies_sorted_file_strm << l << std::endl;
	// NOLINTEND(bugprone-unchecked-optional-access)

	const double median_latency = [&latencies] {
		std::size_t latencies_size = latencies.size();
		std::size_t middle_or_right_index = latencies_size / 2;
		auto middle_or_right_value = static_cast<double>(latencies[middle_or_right_index].count());
		if (latencies_size % 2) {
			return middle_or_right_value;
		}

		auto left = static_cast<double>(latencies[middle_or_right_index - 1].count());
		return (left + middle_or_right_value) / 2.0;
	}();
	auto min_latency = latencies.front();
	auto max_latency = latencies.back();
	auto accumulated_latency =
			std::accumulate(latencies.begin(), latencies.end(), chrn::nanoseconds(0));
	double average_latency =
			static_cast<double>(accumulated_latency.count()) / static_cast<double>(latencies.size());

	// clang-format off
	std::cout << std::endl
						<< "actual-total-work-ns: " << actual_total_work_ns << std::endl
						<< "actual-total-sleep-ns: " << actual_total_sleep_ns << std::endl
						<< "min-latency-ns: " << min_latency.count() << std::endl
						<< "median-latency-ns: " << median_latency << std::endl
						<< "average-latency-ns: " << average_latency << std::endl
						<< "max-latency-ns: " << max_latency.count() << std::endl
	;
	// clang-format on
}

auto main(int argc, char* argv[]) -> int {
	po::options_description desc("Big Fancy Pulser 9000 - Options");
	// clang-format off
	desc.add_options()
		("utilization,u", po::value<double>(&utilization)->default_value(utilization), "The utilization wihtin the range (0, 1.0]")
		("pulse-target-work-ns,p", po::value<std::uint64_t>(&pulse_target_work_ns)->default_value(pulse_target_work_ns), "The target duration in nanoseconds of a single pulse, i.e. a work unit")
		("pulse-concurrency-level,c", po::value<unsigned int>(&pulse_concurrency_level)->default_value(pulse_concurrency_level), "The concurrency level of a single pulse")
		("seed,s", po::value<std::uint_fast64_t>(&seed)->default_value(seed), "The seed of the random number generator")
		("duration-ms,d", po::value<unsigned int>(&duration_ms)->default_value(duration_ms), "The duration of the benchmark in milliseconds")
		("worker-count,w", po::value<workerid_t>(&worker_count)->default_value(worker_count), "The number of worker threads")
		("trace,t", po::bool_switch(&trace)->default_value(trace), "Trace the execution")
		("result-dir,r", po::value(&result_dir), "A directory where to write the evaluation results to")
		("help,h", "Show help")
		;
	// clang-format on

	auto parse_result = po::command_line_parser(argc, argv).options(desc).run();

	po::variables_map vm;
	po::store(parse_result, vm);
	po::notify(vm);

	if (vm.count("help")) {
		std::cout << desc << std::endl;
		return EXIT_SUCCESS;
	}

	try {
		bfp9000();
	} catch (std::exception& e) {
		std::cerr << e.what() << std::endl;
		return EXIT_FAILURE;
	}
}
