// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2022 Florian Fischer
#include <liburing.h>
#include <liburing/io_uring.h>
#include <sys/eventfd.h>
#include <sys/types.h>
#include <unistd.h>

#include <array>
#include <atomic>
#include <cerrno>
#include <chrono>
#include <compare>
#include <cstdint>
#include <functional>
#include <iostream>
#include <string>
#include <thread>

#include "BinaryPrivateSemaphore.hpp"
#include "Common.hpp"
#include "Future.hpp"
#include "Runtime.hpp"
#include "emper-config.h"
#include "lib/math.hpp"

using namespace std::chrono;

static const unsigned ITERATIONS = 10000;
static const unsigned WORK = 1000;

static int evfd;
static uint64_t readBuf = 0;

static std::atomic<uint64_t> end;
static std::array<uint64_t, ITERATIONS> measurements;
static std::atomic<bool> readSubmitted = false;

static void watchFunc() {
	const uint64_t writeBuf = 1;
	for (unsigned i = 0; i < ITERATIONS; ++i) {
		// Clear end time point
		end.store(0, std::memory_order_relaxed);

		// Wait until emper has setup its read request
		readSubmitted.wait(false, std::memory_order_acquire);
		// Consume readSubmitted notification
		readSubmitted.store(false, std::memory_order_relaxed);

		const uint64_t start =
				time_point_cast<nanoseconds>(system_clock::now()).time_since_epoch().count();

		if (write(evfd, &writeBuf, sizeof(writeBuf)) != sizeof(writeBuf))
			DIE_MSG_ERRNO("writing to eventfd failed");

		// while (end.load(std::memory_order_acquire) == 0) {}
		end.wait(0, std::memory_order_acquire);
		measurements[i] = end.load(std::memory_order_relaxed) - start;
	}
}

static void fiberFunc() {
	for (unsigned i = 0; i < ITERATIONS; ++i) {
		emper::io::ReadFuture rf(evfd, &readBuf, sizeof(readBuf), 0);
		BPS bps;

		rf.setCallback([&](const int& res) {
			// Store the time when we are executed
			end.store(time_point_cast<nanoseconds>(system_clock::now()).time_since_epoch().count(),
								std::memory_order_release);

			if (res != sizeof(readBuf) or readBuf != 1) {
				errno = -res;
				DIE_MSG_ERRNO("reading from eventfd failed");
			}

			bps.signal();
			end.notify_one();
		});

		// Submit the read
		rf.submit();

		// Signal the watch that we are setup the read and a measurement can be started
		readSubmitted.store(true, std::memory_order_release);
		readSubmitted.notify_one();

		// Keep the current worker busy
		const auto start = std::chrono::steady_clock::now();
		const auto deadline = start + std::chrono::microseconds(WORK);
		// TODO: The suppressed linter error below may be a false positive
		// reported by clang-tidy.
		// NOLINTNEXTLINE(modernize-use-nullptr)
		while (std::chrono::steady_clock::now() < deadline) {
		}

		// Synchronize with our callback fiber
		bps.wait();
	}
	Runtime::getRuntime()->initiateTermination();
}

static void baselineIoUringFunc() {
	struct io_uring ring;
	int ret = io_uring_queue_init(EMPER_IO_WORKER_URING_ENTRIES, &ring, 0);
	if (ret) DIE_MSG_ERRNO("setting up iouring failed");

	for (unsigned i = 0; i < ITERATIONS; ++i) {
		struct io_uring_sqe* sqe = io_uring_get_sqe(&ring);
		io_uring_prep_read(sqe, evfd, &readBuf, sizeof(readBuf), 0);

		ret = io_uring_submit(&ring);
		if (ret != 1) DIE_MSG_ERRNO("io_uring_submit failed");

		readSubmitted.store(true, std::memory_order_release);
		readSubmitted.notify_one();

		struct io_uring_cqe* cqe;
		ret = io_uring_wait_cqe(&ring, &cqe);
		if (unlikely(!cqe || ret)) DIE_MSG_ERRNO("io_uring_wait_cqe failed");

		end.store(time_point_cast<nanoseconds>(system_clock::now()).time_since_epoch().count(),
							std::memory_order_release);

		if (cqe->res != sizeof(readBuf) || readBuf != 1) DIE_MSG_ERRNO("reading from eventfd failed");

		io_uring_cqe_seen(&ring, cqe);

		end.notify_one();
	}

	io_uring_queue_exit(&ring);
}

static void baselineFunc() {
	for (unsigned i = 0; i < ITERATIONS; ++i) {
		readSubmitted.store(true, std::memory_order_release);
		readSubmitted.notify_one();
		ssize_t res = read(evfd, &readBuf, sizeof(readBuf));

		end.store(time_point_cast<nanoseconds>(system_clock::now()).time_since_epoch().count(),
							std::memory_order_release);

		if (res != sizeof(readBuf) || readBuf != 1) {
			DIE_MSG_ERRNO("reading from eventfd failed");
		}

		end.notify_one();
	}
}

auto main(int argc, char* argv[]) -> int {
	evfd = eventfd(0, 0);
	if (evfd == -1) DIE_MSG_ERRNO("eventfd failed");

	std::string bench = "emper";
	bool printData = false;

	for (int i = 1; i < argc; ++i) {
		auto arg = std::string(argv[i]);
		if (arg == "--data") {
			printData = true;
			continue;
		}

		bench = arg;
		break;
	}

	auto watch = std::thread(watchFunc);

	if (bench == "baseline")
		baselineFunc();
	else if (bench == "iouring")
		baselineIoUringFunc();
	else if (bench == "emper") {
		Runtime runtime;
		runtime.executeAndWait(fiberFunc);
	} else {
		std::cerr << "Usage: " << argv[0] << " [--data] [baseline | iouring]" << std::endl;
		return 1;
	}

	watch.join();
	close(evfd);

	if (printData) {
		for (auto measurement : measurements) std::cout << measurement << std::endl;
	} else {
		emper::lib::math::RunningAverage<uint64_t> latency;
		for (auto measurement : measurements) latency.update(measurement);

		std::cout << "avg latency: " << latency.getAverage() << std::endl;
	}
}
