// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2022 Florian Fischer
#include <boost/program_options.hpp>
#include <chrono>
#include <compare>
#include <cstdint>
#include <cstdlib>
#include <ctime>
#include <exception>
#include <fstream>	// IWYU pragma: keep
#include <functional>
#include <iostream>
#include <string>

#include "CountingPrivateSemaphore.hpp"
#include "Runtime.hpp"
#include "RuntimeBuilder.hpp"
#include "emper-common.h"
#include "emper.hpp"

namespace po = boost::program_options;

using Clock = std::chrono::high_resolution_clock;
using TimePoint = std::chrono::time_point<Clock>;

enum class PulseMode : std::uint8_t {
	Burst, /*!< Introduce all work as fast as possible */
	Even,	 /*!< Introduce work evenly during the pulse */
};

auto operator<<(std::ostream& out, const PulseMode& mode) -> std::ostream& {
	switch (mode) {
		case PulseMode::Burst:
			out << "burst";
			break;
		case PulseMode::Even:
			out << "even";
			break;
	}
	return out;
}

auto operator>>(std::istream& in, PulseMode& mode) -> std::istream& {
	std::string token;

	in >> token;
	if (token == "burst") {
		mode = PulseMode::Burst;
	} else if (token == "even") {
		mode = PulseMode::Even;
	} else {
		in.setstate(std::ios_base::failbit);
	}
	return in;
}

static PulseMode pulseMode;

// Pulse of new work in microseconds
static int64_t pulse = 1L * 1000 * 1000;
// Number of pulses
static unsigned iterations = 30;
// Utilization of the runtime in percent
static unsigned utilization = 80;

class Work {
 public:
	static Work** doneWork;
	static thread_local Work** localDoneWork;

	TimePoint creation;
	TimePoint dispatch;
	Work* next = nullptr;

	Work() : creation(Clock::now()) {}

	void run() {
		this->dispatch = std::chrono::high_resolution_clock::now();
		const auto deadline = Clock::now() + std::chrono::microseconds(pulse);
		while (Clock::now() < deadline) {
		}

		finish();
	}

	void finish() {
		this->next = *localDoneWork;
		*localDoneWork = this;
	}
};

Work** Work::doneWork;
thread_local Work** Work::localDoneWork;

static void pulser(Runtime& runtime) {
	const workerid_t workerCount = runtime.getWorkerCount();
	const workerid_t workItems = (workerCount * utilization) / 100;

	struct timespec ts {
		0, 0
	};
	// TODO: better calculate the time until the next burst
	if (pulseMode == PulseMode::Burst)
		ts.tv_nsec = pulse * 1000;
	else
		ts.tv_nsec = (pulse * 1000) / workItems;

	CPS cps;
	for (unsigned i = 0; i < iterations; ++i) {
		for (unsigned w = 0; w < workItems; ++w) {
			Work* work = new Work();
			spawn([=] { work->run(); }, cps);
			if (pulseMode == PulseMode::Even) emper::nanosleep(&ts);
		}

		// TODO: Think about busy looping for small times
		if (pulseMode == PulseMode::Burst) emper::nanosleep(&ts);
	}
	cps.wait();
	runtime.initiateTermination();
}

static auto eval(const po::variables_map& vm) -> int {
	std::cerr << "Starting pulse evaluation with pulse=" << pulse;
	std::cerr << ", iterations=" << iterations;
	std::cerr << " and utilization=" << utilization;
	std::cerr << std::endl;

	const workerid_t workerCount = Runtime::getDefaultWorkerCount();
	Work::doneWork = new Work*[workerCount];

	RuntimeBuilder runtimeBuilder;
	Runtime runtime = runtimeBuilder.withWorkerCount(workerCount)
												.newWorkerHook([](workerid_t id) {
													Work::localDoneWork = &Work::doneWork[id];
													*Work::localDoneWork = nullptr;
												})
												.build();

	runtime.executeAndWait([&] { pulser(runtime); });

	std::fstream fout;
	std::ostream* out = &std::cout;
	if (vm.count("latencies-file")) {
		out = &fout;
		fout.open(vm["latencies-file"].as<std::string>(), std::fstream::out);
	}

	for (unsigned i = 0; i < workerCount; ++i) {
		auto* localDoneWork = Work::doneWork[i];
		while (localDoneWork) {
			auto* work = localDoneWork;
			*out << i << ","
					 << std::chrono::duration_cast<std::chrono::nanoseconds>(work->dispatch - work->creation)
									.count()
					 << std::endl;
			localDoneWork = work->next;
			delete work;
		}
	}
	delete[] Work::doneWork;

	if (fout.is_open()) fout.close();

	return EXIT_SUCCESS;
}

auto main(int argc, char* argv[]) -> int {
	po::options_description desc("Allowed options");
	// clang-format off
	desc.add_options()
		("help", "Show help")
		("pulse", po::value<int64_t>(&pulse)->default_value(pulse), "The pulse of new work")
		("iterations", po::value<unsigned>(&iterations)->default_value(iterations),
									 "The number of pulses")
		("utilization", po::value<unsigned>(&utilization)->default_value(utilization),
										"The target utilization of the runtime")
		("latencies-file", po::value<std::string>(), "File to store all collected latencies")
		("pulse-mode", po::value<PulseMode>(&pulseMode)->default_value(PulseMode::Burst), "How to introducing the work")
		;
	// clang-format on

	auto parse_result = po::command_line_parser(argc, argv).options(desc).run();

	po::variables_map vm;
	po::store(parse_result, vm);
	po::notify(vm);

	if (vm.count("help")) {
		std::cout << desc << "\n";
		return EXIT_SUCCESS;
	}

	try {
		return eval(vm);
	} catch (std::exception& e) {
		std::cerr << e.what() << std::endl;
		return EXIT_FAILURE;
	}
}
