// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#include <sys/eventfd.h>	// for eventfd, EFD_SEMAPHORE

#include <cerrno>		// for ECANCELED, ETIME
#include <cstdint>	// for uint64_t, int32_t
#include <cstdlib>
#include <ostream>

#include "Common.hpp"
#include "CountingPrivateSemaphore.hpp"
#include "Emper.hpp"
#include "Runtime.hpp"
#include "Semaphore.hpp"
#include "emper.hpp"
#include "fixtures/assert.hpp"
#include "io.hpp"
#include "io/Future.hpp"
#include "lib/LinuxVersion.hpp"

using emper::io::ReadFuture;
using emper::io::WriteFuture;

int efd;
uint64_t read_buf;
uint64_t write_buf = 42;

static void cancelSubmittedCallback() {
	ReadFuture readFuture(efd, &read_buf, sizeof(read_buf), 0);
	readFuture.setCallback([](int32_t res) {
		ASSERT_MSG(res == -ECANCELED, "Expected res == -ECANCELED, got " << res);
	});
	readFuture.submit();
	int32_t res = readFuture.cancel();

	if constexpr (emper::IO_SINGLE_URING) {
		// When using a single io_uring it is possible the submitted Future was
		// canceled before reaching the io_uring.
		ASSERT(!res || res == -ECANCELED);
	} else {
		ASSERT(!res);
	}
}

static void cancelNotSubmitted() {
	ReadFuture readFuture(efd, &read_buf, sizeof(read_buf), 0);
	ASSERT(readFuture.cancel() == -ENOENT);
}

static void cancelSubmittedNotCompleted() {
	ReadFuture readFuture(efd, &read_buf, sizeof(read_buf), 0);
	readFuture.submit();
	ASSERT(readFuture.cancel() == -ECANCELED);
}

static void cancelCompleted() {
	ReadFuture readFuture(efd, &read_buf, sizeof(read_buf), 0);
	readFuture.submit();
	WriteFuture writeFuture(efd, &write_buf, sizeof(write_buf), 0);
	ASSERT(writeFuture.submitAndWait() == sizeof(write_buf));
	ASSERT(readFuture.cancel() == sizeof(write_buf) && read_buf == write_buf);
}

static void cancelNotCompletedChain() {
	ReadFuture readFuture(efd, &read_buf, sizeof(read_buf), 0);
	ReadFuture readFuture2(efd, &read_buf, sizeof(read_buf), 0);
	readFuture2.setDependency(readFuture);

	readFuture2.submit();
	ASSERT(readFuture2.cancel() == -ECANCELED);
	ASSERT(readFuture.get() == -ECANCELED);
}

static void cancelPartialCompletedChain() {
	ReadFuture readFuture(efd, &read_buf, sizeof(read_buf), 0);
	WriteFuture writeFuture(efd, &write_buf, sizeof(write_buf), 0);
	ReadFuture readFuture2(efd, &read_buf, sizeof(read_buf), 0);
	readFuture2.setDependency(readFuture);

	readFuture2.submit();
	ASSERT(writeFuture.submitAndWait() == sizeof(write_buf));
	// TODO: investigate why this read is completed with -EINTR most of the time
	int r = readFuture2.cancel();
	ASSERT(r == -EINTR || r == -ECANCELED);
	ASSERT(readFuture.get() == sizeof(write_buf) && read_buf == write_buf);
}

/**
 * @brief Cancel a lot of read futures hopefully going through scheduleOn
 *
 * The test uses one fiber per worker. Each fiber creates an eventfd, prepares
 * and submits a ReadFuture. Then each Fiber yields and creates a second ReadFuture.
 * After all Fibers are done with this preparation each will issue a write to the
 * next eventfd.
 * Each fiber yields a last time to change the worker and exit resulting in
 * cancellation of two Futures one possible completed and one definitely outstanding.
 */
static void massCancelOnDifferentWorker() {
	const unsigned fiberCount = Runtime::getRuntime()->getWorkerCount() * 5;
	int* evfds = new int[fiberCount];
	emper::Semaphore readySem;
	emper::Semaphore startSem;

	CPS cps;
	for (unsigned i = 0; i < fiberCount; ++i) {
		spawn(
				[&, i = i] {
					uint64_t rbuf = 0;
					const uint64_t wbuf = i;

					evfds[i] = eventfd(0, 0);

					ReadFuture rf(evfds[i], &rbuf, sizeof(rbuf), 0);
					rf.submit();

					emper::yield();

					ReadFuture rf2(evfds[i], &rbuf, sizeof(rbuf), 0);
					rf2.submit();

					readySem.release();
					startSem.acquire();

					WriteFuture wf(evfds[(i + 1) % fiberCount], &wbuf, sizeof(wbuf), 0);
					wf.submitAndWait();

					emper::yield();
					rf.cancel();
					rf2.cancel();
				},
				cps);
	}

	for (unsigned i = 0; i < fiberCount; ++i) {
		readySem.acquire();
	}

	for (unsigned i = 0; i < fiberCount; ++i) {
		startSem.release();
	}
	cps.wait();

	CPS closeCps;
	for (unsigned i = 0; i < fiberCount; ++i) {
		spawn([&, i = i] { emper::io::closeAndForget(evfds[i]); }, closeCps);
	}
	closeCps.wait();
	delete[] evfds;
}

void emperTest() {
	if constexpr (emper::IO_SYNCHRONOUS) exit(77);

	efd = eventfd(0, 0);
	if (efd == -1) {
		DIE_MSG_ERRNO("eventfd failed");
	}

	cancelSubmittedCallback();
	cancelNotSubmitted();
	cancelSubmittedNotCompleted();
	cancelCompleted();
	cancelNotCompletedChain();
	cancelPartialCompletedChain();

	// Async cancelation is racy before linux 5.17.
	// Work may be not found at all or only the work is cancelled but
	// not the armed poll.
	// https://lore.kernel.org/io-uring/20220119024241.609233-1-axboe@kernel.dk/T/#t
	if (EMPER_LINUX_GE("5.17.0")) {
		massCancelOnDifferentWorker();
	}
}
